from __future__ import annotations

from typing import Any, Callable, Literal, Mapping, Sequence

import pygame

from ..entities import (
    Car,
    EmptyFuelCan,
    Flashlight,
    FuelCan,
    FuelStation,
    PatrolBot,
    Player,
    Shoes,
    SpikyPlant,
    Survivor,
    Zombie,
    ZombieDog,
    random_position_outside_building,
    spritecollideany_walls,
)
from ..entities.zombie_movement import (
    _zombie_lineformer_train_head_movement,
    _zombie_solitary_movement,
)
from ..entities_constants import (
    FAST_ZOMBIE_BASE_SPEED,
    PLAYER_SPEED,
    ZombieKind,
    ZOMBIE_DECAY_DURATION_FRAMES,
    ZOMBIE_SPEED,
)
from ..gameplay_constants import (
    DEFAULT_FLASHLIGHT_SPAWN_COUNT,
    DEFAULT_SHOES_SPAWN_COUNT,
)
from ..level_constants import DEFAULT_CELL_SIZE, DEFAULT_GRID_COLS, DEFAULT_GRID_ROWS
from ..models import DustRing, FallingEntity, GameData, Stage
from ..rng import get_rng
from .constants import (
    FALLING_ZOMBIE_DUST_DURATION_MS,
    FALLING_ZOMBIE_DURATION_MS,
    FALLING_ZOMBIE_PRE_FX_MS,
    LAYER_HOUSEPLANTS,
    LAYER_PLAYERS,
    LAYER_VEHICLES,
    LAYER_ZOMBIES,
    MAX_ZOMBIES,
    ZOMBIE_SPAWN_PLAYER_BUFFER,
    ZOMBIE_TRACKER_DECAY_DURATION_FRAMES,
)
from .utils import (
    fov_radius_for_flashlights,
    find_exterior_spawn_position,
    find_interior_spawn_positions,
    find_nearby_offscreen_spawn_position,
    rect_visible_on_screen,
)

RNG = get_rng()

FallScheduleResult = Literal["scheduled", "no_position", "blocked", "no_player"]

__all__ = [
    "place_new_car",
    "place_fuel_can",
    "place_empty_fuel_can",
    "place_fuel_station",
    "place_flashlights",
    "place_shoes",
    "place_buddies",
    "spawn_survivors",
    "setup_player_and_cars",
    "spawn_initial_zombies",
    "spawn_initial_patrol_bots",
    "spawn_spiky_plants",
    "spawn_waiting_car",
    "maintain_waiting_car_supply",
    "nearest_waiting_car",
    "update_falling_zombies",
    "spawn_exterior_zombie",
    "spawn_weighted_zombie",
]


def _cell_center(cell: tuple[int, int], cell_size: int) -> tuple[int, int]:
    return (
        int((cell[0] * cell_size) + (cell_size / 2)),
        int((cell[1] * cell_size) + (cell_size / 2)),
    )


def _car_appearance_for_stage(stage: Stage | None) -> str:
    return "disabled" if stage and stage.endurance_stage else "default"


def _pick_zombie_variant(stage: Stage | None) -> ZombieKind:
    normal_ratio = 1.0
    tracker_ratio = 0.0
    wall_hugging_ratio = 0.0
    lineformer_ratio = 0.0
    solitary_ratio = 0.0
    dog_ratio = 0.0
    if stage is not None:
        normal_ratio = max(0.0, float(stage.zombie_normal_ratio))
        tracker_ratio = max(0.0, float(stage.zombie_tracker_ratio))
        wall_hugging_ratio = max(0.0, float(stage.zombie_wall_hugging_ratio))
        lineformer_ratio = max(0.0, float(stage.zombie_lineformer_ratio))
        solitary_ratio = max(0.0, float(stage.zombie_solitary_ratio))
        # Nimble dogs are a dog sub-variant; include them in the dog weight so
        # total zombie-type weights can always be normalized.
        nimble_ratio = max(0.0, float(stage.zombie_nimble_dog_ratio))
        dog_ratio = max(0.0, float(stage.zombie_dog_ratio)) + nimble_ratio
    total_ratio = (
        normal_ratio
        + tracker_ratio
        + wall_hugging_ratio
        + lineformer_ratio
        + solitary_ratio
        + dog_ratio
    )
    if total_ratio <= 0:
        return ZombieKind.NORMAL
    pick = RNG.random() * total_ratio
    if pick < normal_ratio:
        return ZombieKind.NORMAL
    if pick < normal_ratio + tracker_ratio:
        return ZombieKind.TRACKER
    if pick < normal_ratio + tracker_ratio + wall_hugging_ratio:
        return ZombieKind.WALL_HUGGER
    if pick < normal_ratio + tracker_ratio + wall_hugging_ratio + lineformer_ratio:
        return ZombieKind.LINEFORMER
    if pick < (
        normal_ratio
        + tracker_ratio
        + wall_hugging_ratio
        + lineformer_ratio
        + solitary_ratio
    ):
        return ZombieKind.SOLITARY
    return ZombieKind.DOG


def _build_initial_zombie_kind_plan(
    stage: Stage | None, total: int
) -> list[ZombieKind]:
    if total <= 0:
        return []
    normal_ratio = 1.0
    tracker_ratio = 0.0
    wall_hugging_ratio = 0.0
    lineformer_ratio = 0.0
    solitary_ratio = 0.0
    dog_ratio = 0.0
    if stage is not None:
        normal_ratio = max(0.0, float(stage.zombie_normal_ratio))
        tracker_ratio = max(0.0, float(stage.zombie_tracker_ratio))
        wall_hugging_ratio = max(0.0, float(stage.zombie_wall_hugging_ratio))
        lineformer_ratio = max(0.0, float(stage.zombie_lineformer_ratio))
        solitary_ratio = max(0.0, float(stage.zombie_solitary_ratio))
        nimble_ratio = max(0.0, float(stage.zombie_nimble_dog_ratio))
        dog_ratio = max(0.0, float(stage.zombie_dog_ratio)) + nimble_ratio
    weighted_kinds = [
        (ZombieKind.NORMAL, normal_ratio),
        (ZombieKind.TRACKER, tracker_ratio),
        (ZombieKind.WALL_HUGGER, wall_hugging_ratio),
        (ZombieKind.LINEFORMER, lineformer_ratio),
        (ZombieKind.SOLITARY, solitary_ratio),
        (ZombieKind.DOG, dog_ratio),
    ]
    total_ratio = sum(weight for _, weight in weighted_kinds)
    if total_ratio <= 0.0:
        return [ZombieKind.NORMAL] * total
    planned: list[ZombieKind] = []
    for kind, weight in weighted_kinds:
        if weight <= 0.0:
            continue
        count = int(round(total * (weight / total_ratio)))
        if count == 0:
            count = 1
        planned.extend([kind] * count)
    if not planned:
        return [ZombieKind.NORMAL] * total
    RNG.shuffle(planned)
    if len(planned) >= total:
        return planned[:total]
    while len(planned) < total:
        planned.append(RNG.choice(planned))
    RNG.shuffle(planned)
    return planned


def _is_spawn_position_clear(
    game_data: GameData,
    candidate: Zombie | ZombieDog,
    *,
    allow_player_overlap: bool = False,
) -> bool:
    wall_group = game_data.groups.wall_group
    if spritecollideany_walls(candidate, wall_group):
        return False

    spawn_rect = candidate.rect
    player = game_data.player
    if not allow_player_overlap and player and spawn_rect.colliderect(player.rect):
        return False
    car = game_data.car
    if car and car.alive() and spawn_rect.colliderect(car.rect):
        return False
    for parked in game_data.waiting_cars:
        if parked.alive() and spawn_rect.colliderect(parked.rect):
            return False
    for survivor in game_data.groups.survivor_group:
        if survivor.alive() and spawn_rect.colliderect(survivor.rect):
            return False
    for zombie in game_data.groups.zombie_group:
        if zombie.alive() and spawn_rect.colliderect(zombie.rect):
            return False
    for bot in game_data.groups.patrol_bot_group:
        if bot.alive() and spawn_rect.colliderect(bot.rect):
            return False
    return True


def _is_patrol_spawn_position_clear(
    game_data: GameData,
    candidate: PatrolBot,
    *,
    allow_player_overlap: bool = False,
) -> bool:
    wall_group = game_data.groups.wall_group
    if spritecollideany_walls(candidate, wall_group):
        return False

    spawn_rect = candidate.rect
    player = game_data.player
    if not allow_player_overlap and player and spawn_rect.colliderect(player.rect):
        return False
    car = game_data.car
    if car and car.alive() and spawn_rect.colliderect(car.rect):
        return False
    for parked in game_data.waiting_cars:
        if parked.alive() and spawn_rect.colliderect(parked.rect):
            return False
    for survivor in game_data.groups.survivor_group:
        if survivor.alive() and spawn_rect.colliderect(survivor.rect):
            return False
    for zombie in game_data.groups.zombie_group:
        if zombie.alive() and spawn_rect.colliderect(zombie.rect):
            return False
    for bot in game_data.groups.patrol_bot_group:
        if bot.alive() and spawn_rect.colliderect(bot.rect):
            return False
    return True


def _pick_fall_spawn_position(
    game_data: GameData,
    *,
    min_distance: float,
    attempts: int = 10,
    is_clear: Callable[[tuple[int, int]], bool] | None = None,
) -> tuple[int, int] | None:
    player = game_data.player
    if not player:
        return None
    fall_spawn_cells = game_data.layout.fall_spawn_cells
    if not fall_spawn_cells:
        return None
    car = game_data.car
    target_sprite = car if player.in_car and car and car.alive() else player
    target_center = target_sprite.rect.center
    cell_size = game_data.cell_size
    fov_radius = fov_radius_for_flashlights(game_data.state.flashlight_count)
    min_dist_sq = min_distance * min_distance
    max_dist_sq = fov_radius * fov_radius
    wall_cells = game_data.layout.wall_cells

    candidates: list[tuple[int, int]] = []
    for cell_x, cell_y in fall_spawn_cells:
        if (cell_x, cell_y) in wall_cells:
            continue
        pos = (
            int(cell_x * cell_size + cell_size // 2),
            int(cell_y * cell_size + cell_size // 2),
        )
        dx = pos[0] - target_center[0]
        dy = pos[1] - target_center[1]
        dist_sq = dx * dx + dy * dy
        if dist_sq < min_dist_sq or dist_sq > max_dist_sq:
            continue
        candidates.append(pos)

    if not candidates:
        return None

    RNG.shuffle(candidates)
    for pos in candidates[: max(1, min(attempts, len(candidates)))]:
        if is_clear is not None and not is_clear(pos):
            continue
        return pos
    return None


def _schedule_falling_zombie(
    game_data: GameData,
    config: dict[str, Any],
    *,
    allow_carry: bool = True,
) -> FallScheduleResult:
    player = game_data.player
    if not player:
        return "no_player"
    state = game_data.state
    zombie_group = game_data.groups.zombie_group
    if len(zombie_group) + len(state.falling_zombies) >= MAX_ZOMBIES:
        return "blocked"
    min_distance = game_data.stage.cell_size * 0.5
    kind = _pick_zombie_variant(game_data.stage)

    def _candidate_clear(pos: tuple[int, int]) -> bool:
        candidate = _create_zombie(
            config,
            start_pos=pos,
            stage=game_data.stage,
            kind=kind,
        )
        return _is_spawn_position_clear(game_data, candidate)

    spawn_pos = _pick_fall_spawn_position(
        game_data,
        min_distance=min_distance,
        is_clear=_candidate_clear,
    )
    if spawn_pos is None:
        if allow_carry:
            state.falling_spawn_carry += 1
        return "no_position"
    # start_offset removed; animation handles "falling" via scaling now.
    start_pos = (int(spawn_pos[0]), int(spawn_pos[1]))
    fall = FallingEntity(
        start_pos=start_pos,
        target_pos=(int(spawn_pos[0]), int(spawn_pos[1])),
        started_at_ms=game_data.state.clock.elapsed_ms,
        pre_fx_ms=FALLING_ZOMBIE_PRE_FX_MS,
        fall_duration_ms=FALLING_ZOMBIE_DURATION_MS,
        dust_duration_ms=FALLING_ZOMBIE_DUST_DURATION_MS,
        kind=kind,
    )
    state.falling_zombies.append(fall)
    return "scheduled"


def _create_zombie(
    config: dict[str, Any],
    *,
    start_pos: tuple[int, int] | None = None,
    hint_pos: tuple[float, float] | None = None,
    stage: Stage | None = None,
    kind: ZombieKind | None = None,
) -> Zombie | ZombieDog:
    """Factory to create zombies with optional fast variants."""
    fast_conf = config.get("fast_zombies", {})
    fast_enabled = fast_conf.get("enabled", True)
    if fast_enabled:
        base_speed = RNG.uniform(ZOMBIE_SPEED, FAST_ZOMBIE_BASE_SPEED)
    else:
        base_speed = ZOMBIE_SPEED
    if stage is not None:
        decay_duration_frames = max(
            1.0,
            float(stage.zombie_decay_duration_frames),
        )
    else:
        decay_duration_frames = ZOMBIE_DECAY_DURATION_FRAMES
    if kind is None:
        kind = _pick_zombie_variant(stage)
    base_speed = min(base_speed, PLAYER_SPEED - 0.05)
    if kind == ZombieKind.TRACKER:
        ratio = (
            ZOMBIE_TRACKER_DECAY_DURATION_FRAMES / ZOMBIE_DECAY_DURATION_FRAMES
            if ZOMBIE_DECAY_DURATION_FRAMES > 0
            else 1.0
        )
        decay_duration_frames = max(1.0, decay_duration_frames * ratio)
    if start_pos is None:
        cell_size = stage.cell_size if stage else DEFAULT_CELL_SIZE
        if stage is None:
            grid_cols = DEFAULT_GRID_COLS
            grid_rows = DEFAULT_GRID_ROWS
        else:
            grid_cols = stage.grid_cols
            grid_rows = stage.grid_rows
        level_width = grid_cols * cell_size
        level_height = grid_rows * cell_size
        if hint_pos is not None:
            points = [
                random_position_outside_building(level_width, level_height)
                for _ in range(5)
            ]
            points.sort(
                key=lambda p: (p[0] - hint_pos[0]) ** 2 + (p[1] - hint_pos[1]) ** 2
            )
            start_pos = points[0]
        else:
            start_pos = random_position_outside_building(level_width, level_height)
    if kind == ZombieKind.DOG:
        nimble_ratio = 0.0
        if stage is not None:
            dog_weight = max(0.0, float(stage.zombie_dog_ratio))
            nimble_weight = max(0.0, float(stage.zombie_nimble_dog_ratio))
            total_dog_weight = dog_weight + nimble_weight
            if total_dog_weight > 0.0:
                nimble_ratio = nimble_weight / total_dog_weight
        return ZombieDog(
            x=float(start_pos[0]),
            y=float(start_pos[1]),
            variant="nimble" if RNG.random() < nimble_ratio else "normal",
        )
    movement_strategy = None
    if kind == ZombieKind.LINEFORMER:
        movement_strategy = _zombie_lineformer_train_head_movement
    elif kind == ZombieKind.SOLITARY:
        movement_strategy = _zombie_solitary_movement
    return Zombie(
        x=float(start_pos[0]),
        y=float(start_pos[1]),
        speed=base_speed,
        kind=kind,
        movement_strategy=movement_strategy,
        decay_duration_frames=decay_duration_frames,
    )


def _spawn_lineformer_request(
    game_data: GameData,
    config: dict[str, Any],
    *,
    start_pos: tuple[int, int],
    allow_player_overlap: bool = False,
    check_walls: bool = True,
) -> Zombie | None:
    zombie_group = game_data.groups.zombie_group
    all_sprites = game_data.groups.all_sprites
    wall_group = game_data.groups.wall_group
    manager = game_data.lineformer_trains
    train_id, target_id = manager.resolve_spawn_target(zombie_group, start_pos)
    if train_id is not None:
        if manager.append_marker(train_id, start_pos):
            return manager.get_train_head(train_id, zombie_group)
        return None
    candidate = _create_zombie(
        config,
        start_pos=start_pos,
        stage=game_data.stage,
        kind=ZombieKind.LINEFORMER,
    )
    if not isinstance(candidate, Zombie):
        return None
    if check_walls and spritecollideany_walls(candidate, wall_group):
        return None
    if not _is_spawn_position_clear(
        game_data,
        candidate,
        allow_player_overlap=allow_player_overlap,
    ):
        return None
    zombie_group.add(candidate)
    all_sprites.add(candidate, layer=LAYER_ZOMBIES)
    manager.create_train_for_head(
        candidate,
        target_id=target_id,
        now_ms=game_data.state.clock.elapsed_ms,
    )
    return candidate


def place_fuel_can(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
    count: int = 1,
) -> FuelCan | None:
    """Pick a spawn spot for the fuel can away from the player (and car if given)."""
    return _place_collectible(
        walkable_cells,
        cell_size,
        player,
        cars=cars,
        reserved_centers=reserved_centers,
        count=count,
        min_player_dist=250,
        min_car_dist=200,
        factory=FuelCan,
    )


def place_empty_fuel_can(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
    count: int = 1,
) -> EmptyFuelCan | None:
    """Pick a spawn spot for the empty fuel can away from the player/car."""
    placed = _place_collectible(
        walkable_cells,
        cell_size,
        player,
        cars=cars,
        reserved_centers=reserved_centers,
        count=count,
        min_player_dist=250,
        min_car_dist=200,
        factory=EmptyFuelCan,
    )
    return placed if isinstance(placed, EmptyFuelCan) else None


def place_fuel_station(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
    count: int = 1,
) -> FuelStation | None:
    """Pick a spawn spot for the fuel station away from the player/car."""
    placed = _place_collectible(
        walkable_cells,
        cell_size,
        player,
        cars=cars,
        reserved_centers=reserved_centers,
        count=count,
        min_player_dist=260,
        min_car_dist=200,
        factory=FuelStation,
    )
    return placed if isinstance(placed, FuelStation) else None


def _place_collectible(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None,
    reserved_centers: set[tuple[int, int]] | None,
    count: int,
    min_player_dist: int,
    min_car_dist: int,
    factory: Callable[[int, int], pygame.sprite.Sprite],
) -> pygame.sprite.Sprite | None:
    if count <= 0 or not walkable_cells:
        return None

    min_player_dist_sq = min_player_dist * min_player_dist
    min_car_dist_sq = min_car_dist * min_car_dist

    for _ in range(200):
        cell = RNG.choice(walkable_cells)
        center = _cell_center(cell, cell_size)
        if reserved_centers and center in reserved_centers:
            continue
        dx = center[0] - player.x
        dy = center[1] - player.y
        if dx * dx + dy * dy < min_player_dist_sq:
            continue
        if cars:
            too_close = False
            for parked_car in cars:
                dx = center[0] - parked_car.rect.centerx
                dy = center[1] - parked_car.rect.centery
                if dx * dx + dy * dy < min_car_dist_sq:
                    too_close = True
                    break
            if too_close:
                continue
        return factory(center[0], center[1])

    cell = RNG.choice(walkable_cells)
    center = _cell_center(cell, cell_size)
    return factory(center[0], center[1])


def _place_flashlight(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
) -> Flashlight | None:
    """Pick a spawn spot for the flashlight away from the player (and car if given)."""
    if not walkable_cells:
        return None

    min_player_dist = 260
    min_car_dist = 200
    min_player_dist_sq = min_player_dist * min_player_dist
    min_car_dist_sq = min_car_dist * min_car_dist

    for _ in range(200):
        cell = RNG.choice(walkable_cells)
        center = _cell_center(cell, cell_size)
        if reserved_centers and center in reserved_centers:
            continue
        dx = center[0] - player.x
        dy = center[1] - player.y
        if dx * dx + dy * dy < min_player_dist_sq:
            continue
        if cars:
            if any(
                (center[0] - parked.rect.centerx) ** 2
                + (center[1] - parked.rect.centery) ** 2
                < min_car_dist_sq
                for parked in cars
            ):
                continue
        return Flashlight(center[0], center[1])

    cell = RNG.choice(walkable_cells)
    center = _cell_center(cell, cell_size)
    return Flashlight(center[0], center[1])


def place_flashlights(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
    count: int = DEFAULT_FLASHLIGHT_SPAWN_COUNT,
) -> list[Flashlight]:
    """Spawn flashlights from blueprint cells (with fallback sampling if needed)."""
    if count <= 0 or not walkable_cells:
        return []

    # Blueprint-provided flashlight cells are explicit spawn points.
    placed: list[Flashlight] = []
    for cell in walkable_cells:
        if len(placed) >= count:
            break
        center = _cell_center(cell, cell_size)
        if reserved_centers and center in reserved_centers:
            continue
        placed.append(Flashlight(center[0], center[1]))
    if len(placed) >= count:
        return placed

    # Fallback for legacy/random candidate lists.
    attempts = 0
    max_attempts = max(200, count * 80)
    while len(placed) < count and attempts < max_attempts:
        attempts += 1
        fl = _place_flashlight(
            walkable_cells,
            cell_size,
            player,
            cars=cars,
            reserved_centers=reserved_centers,
        )
        if not fl:
            break
        # Avoid clustering too tightly
        if any(
            (other.rect.centerx - fl.rect.centerx) ** 2
            + (other.rect.centery - fl.rect.centery) ** 2
            < 120 * 120
            for other in placed
        ):
            continue
        placed.append(fl)
    return placed


def _place_shoes(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
) -> Shoes | None:
    """Pick a spawn spot for the shoes away from the player (and car if given)."""
    if not walkable_cells:
        return None

    min_player_dist = 240
    min_car_dist = 200
    min_player_dist_sq = min_player_dist * min_player_dist
    min_car_dist_sq = min_car_dist * min_car_dist

    for _ in range(200):
        cell = RNG.choice(walkable_cells)
        center = _cell_center(cell, cell_size)
        if reserved_centers and center in reserved_centers:
            continue
        dx = center[0] - player.x
        dy = center[1] - player.y
        if dx * dx + dy * dy < min_player_dist_sq:
            continue
        if cars:
            if any(
                (center[0] - parked.rect.centerx) ** 2
                + (center[1] - parked.rect.centery) ** 2
                < min_car_dist_sq
                for parked in cars
            ):
                continue
        return Shoes(center[0], center[1])

    cell = RNG.choice(walkable_cells)
    center = _cell_center(cell, cell_size)
    return Shoes(center[0], center[1])


def place_shoes(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    cars: Sequence[Car] | None = None,
    reserved_centers: set[tuple[int, int]] | None = None,
    count: int = DEFAULT_SHOES_SPAWN_COUNT,
) -> list[Shoes]:
    """Spawn multiple shoes using the single-place helper to spread them out."""
    placed: list[Shoes] = []
    attempts = 0
    max_attempts = max(200, count * 80)
    while len(placed) < count and attempts < max_attempts:
        attempts += 1
        shoes = _place_shoes(
            walkable_cells,
            cell_size,
            player,
            cars=cars,
            reserved_centers=reserved_centers,
        )
        if not shoes:
            break
        if any(
            (other.rect.centerx - shoes.rect.centerx) ** 2
            + (other.rect.centery - shoes.rect.centery) ** 2
            < 120 * 120
            for other in placed
        ):
            continue
        placed.append(shoes)
    return placed


def place_buddies(
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    player: Player,
    *,
    count: int = 1,
) -> list[Survivor]:
    placed: list[Survivor] = []
    if count <= 0 or not walkable_cells:
        return placed
    min_player_dist = 240
    positions = find_interior_spawn_positions(
        walkable_cells,
        cell_size,
        1.0,
        player=player,
        min_player_dist=min_player_dist,
    )
    RNG.shuffle(positions)
    for pos in positions[:count]:
        placed.append(Survivor(pos[0], pos[1], is_buddy=True))
    remaining = count - len(placed)
    for _ in range(max(0, remaining)):
        spawn_pos = find_nearby_offscreen_spawn_position(
            walkable_cells,
            cell_size,
        )
        placed.append(Survivor(spawn_pos[0], spawn_pos[1], is_buddy=True))
    return placed


def place_new_car(
    wall_group: pygame.sprite.Group,
    player: Player,
    walkable_cells: list[tuple[int, int]],
    cell_size: int,
    *,
    existing_cars: Sequence[Car] | None = None,
    appearance: str = "default",
) -> Car | None:
    if not walkable_cells:
        return None

    max_attempts = 150
    for _ in range(max_attempts):
        cell = RNG.choice(walkable_cells)
        c_x, c_y = _cell_center(cell, cell_size)
        temp_car = Car(c_x, c_y, appearance=appearance)
        temp_rect = temp_car.rect.inflate(30, 30)
        nearby_walls = pygame.sprite.Group()
        nearby_walls.add(
            [
                w
                for w in wall_group
                if abs(w.rect.centerx - c_x) < 150 and abs(w.rect.centery - c_y) < 150
            ]
        )
        collides_wall = spritecollideany_walls(temp_car, nearby_walls)
        collides_player = temp_rect.colliderect(player.rect.inflate(50, 50))
        car_overlap = False
        if existing_cars:
            car_overlap = any(
                temp_car.rect.colliderect(other.rect)
                for other in existing_cars
                if other and other.alive()
            )
        if not collides_wall and not collides_player and not car_overlap:
            return temp_car
    return None


def spawn_survivors(
    game_data: GameData, layout_data: Mapping[str, list[tuple[int, int]]]
) -> list[Survivor]:
    """Populate rescue-stage survivors and buddy-stage buddies."""
    survivors: list[Survivor] = []
    spawn_rate = max(0.0, game_data.stage.survivor_spawn_rate)
    if not (spawn_rate > 0.0 or game_data.stage.buddy_required_count > 0):
        return survivors

    walkable = layout_data.get("walkable_cells", [])
    wall_group = game_data.groups.wall_group
    survivor_group = game_data.groups.survivor_group
    all_sprites = game_data.groups.all_sprites
    cell_size = game_data.cell_size

    if spawn_rate > 0.0:
        positions = find_interior_spawn_positions(
            walkable,
            cell_size,
            spawn_rate,
        )
        for pos in positions:
            survivor = Survivor(*pos)
            if spritecollideany_walls(survivor, wall_group):
                continue
            survivor_group.add(survivor)
            all_sprites.add(survivor, layer=LAYER_PLAYERS)
            survivors.append(survivor)

    if game_data.stage.buddy_required_count > 0:
        buddy_count = max(0, game_data.stage.buddy_required_count)
        buddies: list[Survivor] = []
        if game_data.player:
            buddies = place_buddies(
                walkable,
                cell_size,
                game_data.player,
                count=buddy_count,
            )
        for buddy in buddies:
            if spritecollideany_walls(buddy, wall_group):
                continue
            survivor_group.add(buddy)
            all_sprites.add(buddy, layer=LAYER_PLAYERS)
            survivors.append(buddy)

    return survivors


def setup_player_and_cars(
    game_data: GameData,
    layout_data: Mapping[str, list[tuple[int, int]]],
    *,
    car_count: int = 1,
) -> tuple[Player, list[Car]]:
    """Create the player plus one or more parked cars using blueprint candidates."""
    if not game_data.stage.endurance_stage:
        assert car_count > 0, "Non-endurance stages must have at least one car"

    all_sprites = game_data.groups.all_sprites
    walkable_cells: list[tuple[int, int]] = layout_data["walkable_cells"]
    cell_size = game_data.cell_size
    level_rect = game_data.layout.field_rect

    def _pick_center(cells: list[tuple[int, int]]) -> tuple[int, int]:
        return (
            _cell_center(RNG.choice(cells), cell_size) if cells else level_rect.center
        )

    player_pos = _pick_center(layout_data["player_cells"] or walkable_cells)
    player = Player(*player_pos)

    car_spawn_cells = (
        layout_data.get("car_spawn_cells")
        or layout_data.get("car_walkable_cells")
        or walkable_cells
    )
    spiky_plant_set = set(layout_data.get("spiky_plant_cells", []))
    car_candidates = [
        c
        for c in (layout_data["car_cells"] or car_spawn_cells)
        if c not in spiky_plant_set
    ]
    waiting_cars: list[Car] = []
    car_appearance = _car_appearance_for_stage(game_data.stage)

    def _pick_car_position() -> tuple[int, int]:
        """Favor distant cells for the first car, otherwise fall back to random picks."""
        if not car_candidates:
            return (player_pos[0] + 200, player_pos[1])
        RNG.shuffle(car_candidates)
        for candidate in car_candidates:
            center = _cell_center(candidate, cell_size)
            if (center[0] - player_pos[0]) ** 2 + (
                center[1] - player_pos[1]
            ) ** 2 >= 400 * 400:
                car_candidates.remove(candidate)
                return center
        choice = car_candidates.pop()
        return _cell_center(choice, cell_size)

    for _ in range(max(0, car_count)):
        car_pos = _pick_car_position()
        car = Car(*car_pos, appearance=car_appearance)
        waiting_cars.append(car)
        all_sprites.add(car, layer=LAYER_VEHICLES)
        if not car_candidates:
            break

    all_sprites.add(player, layer=LAYER_PLAYERS)
    return player, waiting_cars


def spawn_initial_zombies(
    game_data: GameData,
    player: Player,
    layout_data: Mapping[str, list[tuple[int, int]]],
    config: dict[str, Any],
) -> None:
    """Spawn initial zombies using blueprint candidate cells."""
    wall_group = game_data.groups.wall_group
    zombie_group = game_data.groups.zombie_group
    all_sprites = game_data.groups.all_sprites

    cell_size = game_data.cell_size
    spawn_cells = layout_data["walkable_cells"]
    if not spawn_cells:
        return

    def _apply_initial_health_gradient(
        zombie: Zombie | ZombieDog,
        *,
        index: int,
        total: int,
        now_ms: int,
    ) -> None:
        if total <= 0:
            return
        if total == 1:
            target_ratio = 1.0
        else:
            progress = index / (total - 1)
            target_ratio = 1.0 - (0.5 * progress)
        max_health = max(1, int(getattr(zombie, "max_health", 1)))
        current_health = max(0, int(getattr(zombie, "health", max_health)))
        target_health = max(1, min(max_health, int(round(max_health * target_ratio))))
        damage = max(0, current_health - target_health)
        if damage <= 0:
            return
        if isinstance(zombie, Zombie):
            zombie.take_damage(damage, source="initial_spawn_gradient", now_ms=now_ms)
        else:
            zombie.take_damage(damage, now_ms=now_ms)

    spawn_rate = max(0.0, game_data.stage.initial_interior_spawn_rate)
    positions = find_interior_spawn_positions(
        spawn_cells,
        cell_size,
        spawn_rate,
        player=player,
        min_player_dist=ZOMBIE_SPAWN_PLAYER_BUFFER,
    )
    kind_plan = _build_initial_zombie_kind_plan(game_data.stage, len(positions))

    for spawn_index, pos in enumerate(positions):
        kind = (
            kind_plan[spawn_index]
            if spawn_index < len(kind_plan)
            else _pick_zombie_variant(game_data.stage)
        )
        if kind == ZombieKind.LINEFORMER:
            _spawn_lineformer_request(
                game_data,
                config,
                start_pos=(int(pos[0]), int(pos[1])),
                check_walls=True,
            )
            continue
        tentative = _create_zombie(
            config,
            start_pos=pos,
            stage=game_data.stage,
            kind=kind,
        )
        if spritecollideany_walls(tentative, wall_group):
            continue
        _apply_initial_health_gradient(
            tentative,
            index=spawn_index,
            total=len(positions),
            now_ms=game_data.state.clock.elapsed_ms,
        )
        zombie_group.add(tentative)
        all_sprites.add(tentative, layer=LAYER_ZOMBIES)

    interval = max(1, game_data.stage.spawn_interval_ms)
    game_data.state.last_zombie_spawn_time = game_data.state.clock.elapsed_ms - interval


def spawn_initial_patrol_bots(
    game_data: GameData,
    player: Player,
    layout_data: Mapping[str, list[tuple[int, int]]],
) -> None:
    """Spawn initial patrol bots using walkable cells and stage spawn rate."""
    spawn_rate = max(0.0, game_data.stage.patrol_bot_spawn_rate)
    if spawn_rate <= 0.0:
        return
    walkable_cells = layout_data.get("walkable_cells", [])
    if not walkable_cells:
        return
    cell_size = game_data.cell_size
    positions = find_interior_spawn_positions(
        walkable_cells,
        cell_size,
        spawn_rate,
        jitter_ratio=0.0,
        player=player,
        min_player_dist=ZOMBIE_SPAWN_PLAYER_BUFFER,
    )
    if not positions:
        return

    patrol_group = game_data.groups.patrol_bot_group
    all_sprites = game_data.groups.all_sprites

    for pos in positions:
        bot = PatrolBot(pos[0], pos[1])
        if not _is_patrol_spawn_position_clear(game_data, bot):
            continue
        patrol_group.add(bot)
        all_sprites.add(bot, layer=LAYER_VEHICLES)


def spawn_waiting_car(game_data: GameData) -> Car | None:
    """Attempt to place an additional parked car on the map."""
    player = game_data.player
    if not player:
        return None
    # Use cells that are 4-way reachable by car (prefer spawn-filtered cells).
    car_spawn_cells = list(game_data.layout.car_spawn_cells)
    if car_spawn_cells:
        walkable_cells = car_spawn_cells
    else:
        car_walkable = list(game_data.layout.car_walkable_cells)
        walkable_cells = (
            car_walkable if car_walkable else game_data.layout.walkable_cells
        )
    if not walkable_cells:
        return None
    wall_group = game_data.groups.wall_group
    all_sprites = game_data.groups.all_sprites
    cell_size = game_data.cell_size
    active_car = game_data.car if game_data.car and game_data.car.alive() else None
    waiting = _alive_waiting_cars(game_data)
    obstacles: list[Car] = list(waiting)
    if active_car:
        obstacles.append(active_car)
    camera = game_data.camera
    appearance = _car_appearance_for_stage(game_data.stage)
    offscreen_attempts = 6
    while offscreen_attempts > 0:
        new_car = place_new_car(
            wall_group,
            player,
            walkable_cells,
            cell_size,
            existing_cars=obstacles,
            appearance=appearance,
        )
        if not new_car:
            return None
        if rect_visible_on_screen(camera, new_car.rect):
            offscreen_attempts -= 1
            continue
        game_data.waiting_cars.append(new_car)
        all_sprites.add(new_car, layer=LAYER_VEHICLES)
        return new_car
    return None


def maintain_waiting_car_supply(
    game_data: GameData, *, minimum: int | None = None
) -> None:
    """Ensure a baseline count of parked cars exists."""
    if minimum is None:
        minimum = game_data.stage.waiting_car_target_count
    target = max(0, minimum)
    current = len(_alive_waiting_cars(game_data))
    while current < target:
        new_car = spawn_waiting_car(game_data)
        if not new_car:
            break
        current += 1


def _alive_waiting_cars(game_data: GameData) -> list[Car]:
    """Return the list of parked cars that still exist, pruning any destroyed sprites."""
    cars = [car for car in game_data.waiting_cars if car.alive()]
    game_data.waiting_cars = cars
    _log_waiting_car_count(game_data)
    return cars


def _log_waiting_car_count(game_data: GameData, *, force: bool = False) -> None:
    """Print the number of waiting cars when it changes."""
    current = len(game_data.waiting_cars)
    if not force and current == game_data.last_logged_waiting_cars:
        return
    game_data.last_logged_waiting_cars = current


def nearest_waiting_car(game_data: GameData, origin: tuple[float, float]) -> Car | None:
    """Find the closest waiting car to an origin point."""
    cars = _alive_waiting_cars(game_data)
    if not cars:
        return None
    return min(
        cars,
        key=lambda car: (
            (car.rect.centerx - origin[0]) ** 2 + (car.rect.centery - origin[1]) ** 2
        ),
    )


def _spawn_nearby_zombie(
    game_data: GameData,
    config: dict[str, Any],
) -> Zombie | None:
    """Spawn a zombie just outside of the current camera frustum."""
    player = game_data.player
    if not player:
        return None
    zombie_group = game_data.groups.zombie_group
    if len(zombie_group) >= MAX_ZOMBIES:
        return None
    camera = game_data.camera
    wall_group = game_data.groups.wall_group
    all_sprites = game_data.groups.all_sprites
    spawn_pos = find_nearby_offscreen_spawn_position(
        game_data.layout.walkable_cells,
        game_data.cell_size,
        player=player,
        camera=camera,
        min_player_dist=ZOMBIE_SPAWN_PLAYER_BUFFER,
        attempts=50,
    )
    kind = _pick_zombie_variant(game_data.stage)
    if kind == ZombieKind.LINEFORMER:
        return _spawn_lineformer_request(
            game_data,
            config,
            start_pos=spawn_pos,
            check_walls=True,
        )
    new_zombie = _create_zombie(
        config,
        start_pos=spawn_pos,
        stage=game_data.stage,
        kind=kind,
    )
    if spritecollideany_walls(new_zombie, wall_group):
        return None
    zombie_group.add(new_zombie)
    all_sprites.add(new_zombie, layer=LAYER_ZOMBIES)
    return new_zombie


def spawn_exterior_zombie(
    game_data: GameData,
    config: dict[str, Any],
) -> Zombie | None:
    """Spawn a zombie using the standard exterior hint logic."""
    player = game_data.player
    if not player:
        return None
    zombie_group = game_data.groups.zombie_group
    all_sprites = game_data.groups.all_sprites
    level_rect = game_data.layout.field_rect
    spawn_pos = find_exterior_spawn_position(
        level_rect.width,
        level_rect.height,
        hint_pos=(player.x, player.y),
    )
    kind = _pick_zombie_variant(game_data.stage)
    if kind == ZombieKind.LINEFORMER:
        return _spawn_lineformer_request(
            game_data,
            config,
            start_pos=spawn_pos,
            check_walls=False,
        )
    new_zombie = _create_zombie(
        config,
        start_pos=spawn_pos,
        stage=game_data.stage,
        kind=kind,
    )
    zombie_group.add(new_zombie)
    all_sprites.add(new_zombie, layer=LAYER_ZOMBIES)
    return new_zombie


def update_falling_zombies(game_data: GameData, config: dict[str, Any]) -> None:
    state = game_data.state
    if not state.falling_zombies:
        return
    now = state.clock.elapsed_ms
    zombie_group = game_data.groups.zombie_group
    all_sprites = game_data.groups.all_sprites
    for fall in list(state.falling_zombies):
        fall_start = fall.started_at_ms + fall.pre_fx_ms
        impact_at = fall_start + fall.fall_duration_ms
        spawn_at = impact_at + fall.dust_duration_ms
        if now >= impact_at and not fall.dust_started:
            state.dust_rings.append(
                DustRing(
                    pos=fall.target_pos,
                    started_at_ms=impact_at,
                    duration_ms=fall.dust_duration_ms,
                )
            )
            fall.dust_started = True
        if now < spawn_at:
            continue

        if getattr(fall, "mode", "spawn") == "spawn":
            if len(zombie_group) < MAX_ZOMBIES:
                if fall.kind == ZombieKind.LINEFORMER:
                    _spawn_lineformer_request(
                        game_data,
                        config,
                        start_pos=fall.target_pos,
                        allow_player_overlap=True,
                        check_walls=False,
                    )
                    state.falling_zombies.remove(fall)
                    continue
                candidate = _create_zombie(
                    config,
                    start_pos=fall.target_pos,
                    stage=game_data.stage,
                    kind=fall.kind,
                )
                zombie_group.add(candidate)
                all_sprites.add(candidate, layer=LAYER_ZOMBIES)

        state.falling_zombies.remove(fall)


def spawn_weighted_zombie(
    game_data: GameData,
    config: dict[str, Any],
) -> bool:
    """Spawn a zombie according to the stage's interior/exterior mix."""
    stage = game_data.stage

    def _spawn_interior() -> bool:
        return _spawn_nearby_zombie(game_data, config) is not None

    def _spawn_exterior() -> bool:
        return spawn_exterior_zombie(game_data, config) is not None

    def _spawn_fall() -> FallScheduleResult:
        result = _schedule_falling_zombie(game_data, config)
        if result != "scheduled":
            return result
        state = game_data.state
        if state.falling_spawn_carry > 0:
            extra = _schedule_falling_zombie(
                game_data,
                config,
                allow_carry=False,
            )
            if extra == "scheduled":
                state.falling_spawn_carry = max(0, state.falling_spawn_carry - 1)
        return "scheduled"

    interior_weight = max(0.0, stage.interior_spawn_weight)
    exterior_weight = max(0.0, stage.exterior_spawn_weight)
    fall_weight = max(0.0, getattr(stage, "interior_fall_spawn_weight", 0.0))
    total_weight = interior_weight + exterior_weight + fall_weight
    if total_weight <= 0:
        # Fall back to exterior spawns if weights are unset or invalid.
        return _spawn_exterior()

    pick = RNG.uniform(0, total_weight)
    if pick <= interior_weight:
        if _spawn_interior():
            return True
        fall_result = _spawn_fall()
        if fall_result == "scheduled":
            return True
        return False
    if pick <= interior_weight + fall_weight:
        fall_result = _spawn_fall()
        if fall_result == "scheduled":
            return True
        return False
    if _spawn_exterior():
        return True
    fall_result = _spawn_fall()
    if fall_result == "scheduled":
        return True
    return False


def spawn_spiky_plants(
    game_data: GameData,
    layout_data: Mapping[str, list[tuple[int, int]]],
) -> list[SpikyPlant]:
    """Spawn spiky plants based on blueprint cells."""
    spiky_plants: list[SpikyPlant] = []
    spiky_plant_cells = layout_data.get("spiky_plant_cells", [])
    if not spiky_plant_cells:
        return spiky_plants

    all_sprites = game_data.groups.all_sprites
    cell_size = game_data.cell_size

    for cell in spiky_plant_cells:
        pos = _cell_center(cell, cell_size)
        spiky_plant = SpikyPlant(pos[0], pos[1])
        all_sprites.add(spiky_plant, layer=LAYER_HOUSEPLANTS)
        spiky_plants.append(spiky_plant)

    return spiky_plants



"""Entity-related constants and helpers."""

from __future__ import annotations

from enum import Enum

from .screen_constants import FPS


class ZombieKind(str, Enum):
    NORMAL = "normal"
    TRACKER = "tracker"
    WALL_HUGGER = "wall_hugging"
    LINEFORMER = "lineformer"
    SOLITARY = "solitary"
    DOG = "dog"


class MovingFloorDirection(str, Enum):
    UP = "U"
    DOWN = "D"
    LEFT = "L"
    RIGHT = "R"


# --- Player and buddy settings ---
HUMANOID_RADIUS = 6
PLAYER_RADIUS = HUMANOID_RADIUS
PLAYER_SPEED = 1.4
MOVING_FLOOR_SPEED = 1.6
FOV_RADIUS = 124  # approximate legacy FOV (80) * 1.55 cap
BUDDY_RADIUS = HUMANOID_RADIUS
BUDDY_FOLLOW_SPEED = PLAYER_SPEED * 0.7
BUDDY_FOLLOW_START_DISTANCE = 30
BUDDY_FOLLOW_STOP_DISTANCE = 100
BUDDY_WALL_DAMAGE_RANGE = 50
BUDDY_MERGE_DISTANCE = 40
HUMANOID_WALL_BUMP_FRAMES = 7
HUMANOID_WALL_BUMP_HOLD_FRAMES = 4
HUMANOID_WALL_BUMP_HOLD_FRAMES = 4

# --- Jump settings ---
PLAYER_JUMP_RANGE = 15  # px (enough to clear one 50px cell)
SURVIVOR_JUMP_RANGE = int(PLAYER_JUMP_RANGE * 0.7)  # px
JUMP_DURATION_MS = 200
JUMP_SCALE_MAX = 0.15
JUMP_SHADOW_OFFSET = 5

# --- Survivor settings (Stage 4) ---
SURVIVOR_RADIUS = HUMANOID_RADIUS
SURVIVOR_APPROACH_RADIUS = 48
SURVIVOR_APPROACH_SPEED = PLAYER_SPEED * 0.35
SURVIVOR_MAX_SAFE_PASSENGERS = 5
SURVIVOR_MIN_SPEED_FACTOR = 0.35

# --- Flashlight settings ---
FLASHLIGHT_WIDTH = 12
FLASHLIGHT_HEIGHT = 10

# --- Shoes settings ---
SHOES_WIDTH = 14
SHOES_HEIGHT = 12

# --- Zombie settings ---
ZOMBIE_RADIUS = HUMANOID_RADIUS
ZOMBIE_SPEED = PLAYER_SPEED * 0.45
ZOMBIE_WANDER_INTERVAL_MS = 5000
ZOMBIE_SIGHT_RANGE = FOV_RADIUS * 1.2
FAST_ZOMBIE_BASE_SPEED = PLAYER_SPEED * 0.77
ZOMBIE_SEPARATION_DISTANCE = ZOMBIE_RADIUS * 2.2
ZOMBIE_DECAY_DURATION_FRAMES = FPS * 60 * 6  # ~6 minutes at target framerate
ZOMBIE_DECAY_MIN_SPEED_RATIO = 0.4
ZOMBIE_TRACKER_SIGHT_RANGE = ZOMBIE_SIGHT_RANGE * 0.2
ZOMBIE_TRACKER_SCENT_RADIUS = 70
ZOMBIE_TRACKER_FAR_SCENT_RADIUS = 140
ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS = 5000
ZOMBIE_TRACKER_SCENT_TOP_K = 3
ZOMBIE_TRACKER_SCAN_INTERVAL_MS = int(1000 * 30 / FPS)
ZOMBIE_TRACKER_WANDER_INTERVAL_MS = 2500
ZOMBIE_TRACKER_CROWD_BAND_WIDTH = 32
ZOMBIE_TRACKER_GRID_CROWD_COUNT = 3
ZOMBIE_TRACKER_RELOCK_DELAY_MS = 3000
ZOMBIE_SOLITARY_SIGHT_RANGE = ZOMBIE_TRACKER_SIGHT_RANGE
ZOMBIE_WALL_HUG_SENSOR_DISTANCE = 30.0
ZOMBIE_WALL_HUG_SENSOR_DIST_RATIO = 0.6
ZOMBIE_WALL_HUG_PROBE_ANGLE_DEG = 55.0
ZOMBIE_WALL_HUG_PROBE_ANGLE_PERP_DEG = 90.0
ZOMBIE_WALL_HUG_TARGET_GAP = 4.2
ZOMBIE_WALL_HUG_LOST_WALL_MS = 2500
ZOMBIE_WALL_HUG_TURN_STEP_DEG = 10.0
ZOMBIE_LINEFORMER_SPEED_MULTIPLIER = 1.4
ZOMBIE_LINEFORMER_JOIN_RADIUS = 50
ZOMBIE_LINEFORMER_FOLLOW_DISTANCE = ZOMBIE_RADIUS * 2.5
ZOMBIE_LINEFORMER_FOLLOW_TOLERANCE = ZOMBIE_RADIUS * 0.6
ZOMBIE_LINEFORMER_DISSOLVE_SPAWN_MS = 120
ZOMBIE_SOLITARY_EVAL_INTERVAL_FRAMES = 10
ZOMBIE_CARBONIZE_DECAY_FRAMES = FPS * 30

# --- Zombie dog settings ---
ZOMBIE_DOG_PATROL_SPEED = PLAYER_SPEED * 0.3
ZOMBIE_DOG_ASSAULT_SPEED = PLAYER_SPEED * 1.6
ZOMBIE_DOG_SIGHT_RANGE = 150
ZOMBIE_DOG_SURVIVOR_SIGHT_RANGE = 50
ZOMBIE_DOG_LONG_AXIS_RATIO = 1.2
ZOMBIE_DOG_SHORT_AXIS_RATIO = 0.6
ZOMBIE_DOG_WANDER_INTERVAL_MS = 1200
ZOMBIE_DOG_PACK_CHASE_RANGE = 50
ZOMBIE_DOG_CHARGE_OFFSET_MIN_NORMAL = 6.0
ZOMBIE_DOG_CHARGE_OFFSET_MAX_NORMAL = 18.0
ZOMBIE_DOG_CHARGE_OFFSET_MIN_NIMBLE = 24.0
ZOMBIE_DOG_CHARGE_OFFSET_MAX_NIMBLE = 60.0
ZOMBIE_DOG_CHARGE_DISTANCE_NORMAL = 480.0
ZOMBIE_DOG_CHARGE_DISTANCE_NIMBLE = 40.0
ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NORMAL = 120
ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NIMBLE = 1200
ZOMBIE_DOG_DECAY_DURATION_FRAMES = max(1, int(ZOMBIE_DECAY_DURATION_FRAMES * 0.5))
ZOMBIE_DOG_DECAY_MIN_SPEED_RATIO = 0.6

# --- Patrol bot settings ---
PATROL_BOT_SPRITE_SIZE = 20
PATROL_BOT_COLLISION_RADIUS = 10.0
PATROL_BOT_DIRECTION_COMMAND_RADIUS = PATROL_BOT_COLLISION_RADIUS * 0.625
PATROL_BOT_DIRECTION_COMMAND_HOLD_MS = 1800
PATROL_BOT_SPEED = PLAYER_SPEED * 0.6
PATROL_BOT_HUMANOID_PAUSE_MS = 500
PATROL_BOT_ZOMBIE_DAMAGE_INTERVAL_FRAMES = 4
PATROL_BOT_PARALYZE_MS = 3000
PATROL_BOT_PARALYZE_BLINK_MS = 200
PATROL_BOT_PARALYZE_MARKER_COLOR = (250, 240, 170)
PATROL_BOT_COLLISION_MARGIN = 4.0

# --- Car and fuel settings ---
CAR_WIDTH = 16
CAR_HEIGHT = 24
CAR_SPEED = 2
CAR_HEALTH = 1000
CAR_WALL_DAMAGE = 20
CAR_PITFALL_FALL_CENTER_RATIO = 0.18
FUEL_CAN_WIDTH = 12
FUEL_CAN_HEIGHT = 15
EMPTY_FUEL_CAN_WIDTH = 12
EMPTY_FUEL_CAN_HEIGHT = 15
FUEL_STATION_WIDTH = 14
FUEL_STATION_HEIGHT = 18

# --- Spiky plant settings ---
SPIKY_PLANT_HEALTH = 300
SPIKY_PLANT_RADIUS = 8
SPIKY_PLANT_COLLISION_RADIUS = 8
SPIKY_PLANT_HUMANOID_SPEED_FACTOR = 0.2
SPIKY_PLANT_CAR_SPEED_FACTOR = 0.2

# --- Puddle settings ---
PUDDLE_SPEED_FACTOR = 0.3
ZOMBIE_CONTAMINATED_SPEED_FACTOR = SPIKY_PLANT_HUMANOID_SPEED_FACTOR
TRAPPED_ZOMBIE_SLOW_FACTOR = SPIKY_PLANT_HUMANOID_SPEED_FACTOR
TRAPPED_ZOMBIE_REPEL_PER_STACK = 0.18
TRAPPED_ZOMBIE_REPEL_MAX_MULT = 1.2
TRAPPED_ZOMBIE_REPEL_RADIUS_CELLS = 1.2

# --- Wall and beam settings ---
INTERNAL_WALL_HEALTH = 40 * 100
INTERNAL_WALL_BEVEL_DEPTH = 6
STEEL_BEAM_HEALTH = int(INTERNAL_WALL_HEALTH * 1.5)
PLAYER_WALL_DAMAGE = 100
BUDDY_WALL_DAMAGE = int(PLAYER_WALL_DAMAGE * 0.7)
ZOMBIE_WALL_DAMAGE = 1
PATROL_BOT_ZOMBIE_DAMAGE = 1

__all__ = [
    "HUMANOID_RADIUS",
    "PLAYER_RADIUS",
    "PLAYER_SPEED",
    "FOV_RADIUS",
    "BUDDY_RADIUS",
    "BUDDY_FOLLOW_SPEED",
    "BUDDY_FOLLOW_START_DISTANCE",
    "BUDDY_FOLLOW_STOP_DISTANCE",
    "BUDDY_WALL_DAMAGE_RANGE",
    "BUDDY_MERGE_DISTANCE",
    "HUMANOID_WALL_BUMP_FRAMES",
    "HUMANOID_WALL_BUMP_HOLD_FRAMES",
    "HUMANOID_WALL_BUMP_HOLD_FRAMES",
    "SURVIVOR_RADIUS",
    "SURVIVOR_APPROACH_RADIUS",
    "SURVIVOR_APPROACH_SPEED",
    "SURVIVOR_MAX_SAFE_PASSENGERS",
    "SURVIVOR_MIN_SPEED_FACTOR",
    "FLASHLIGHT_WIDTH",
    "FLASHLIGHT_HEIGHT",
    "SHOES_WIDTH",
    "SHOES_HEIGHT",
    "ZombieKind",
    "ZOMBIE_RADIUS",
    "ZOMBIE_SPEED",
    "ZOMBIE_WANDER_INTERVAL_MS",
    "ZOMBIE_SIGHT_RANGE",
    "ZOMBIE_TRACKER_SIGHT_RANGE",
    "FAST_ZOMBIE_BASE_SPEED",
    "ZOMBIE_SEPARATION_DISTANCE",
    "ZOMBIE_DECAY_DURATION_FRAMES",
    "ZOMBIE_DECAY_MIN_SPEED_RATIO",
    "ZOMBIE_CARBONIZE_DECAY_FRAMES",
    "ZOMBIE_DOG_DECAY_DURATION_FRAMES",
    "ZOMBIE_DOG_DECAY_MIN_SPEED_RATIO",
    "ZOMBIE_TRACKER_SCENT_RADIUS",
    "ZOMBIE_TRACKER_FAR_SCENT_RADIUS",
    "ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS",
    "ZOMBIE_TRACKER_SCENT_TOP_K",
    "ZOMBIE_TRACKER_SCAN_INTERVAL_MS",
    "ZOMBIE_TRACKER_WANDER_INTERVAL_MS",
    "ZOMBIE_TRACKER_CROWD_BAND_WIDTH",
    "ZOMBIE_TRACKER_GRID_CROWD_COUNT",
    "ZOMBIE_TRACKER_RELOCK_DELAY_MS",
    "ZOMBIE_SOLITARY_SIGHT_RANGE",
    "ZOMBIE_WALL_HUG_SENSOR_DISTANCE",
    "ZOMBIE_WALL_HUG_SENSOR_DIST_RATIO",
    "ZOMBIE_WALL_HUG_PROBE_ANGLE_DEG",
    "ZOMBIE_WALL_HUG_TARGET_GAP",
    "ZOMBIE_WALL_HUG_LOST_WALL_MS",
    "ZOMBIE_LINEFORMER_SPEED_MULTIPLIER",
    "ZOMBIE_LINEFORMER_JOIN_RADIUS",
    "ZOMBIE_LINEFORMER_FOLLOW_DISTANCE",
    "ZOMBIE_LINEFORMER_FOLLOW_TOLERANCE",
    "ZOMBIE_LINEFORMER_DISSOLVE_SPAWN_MS",
    "ZOMBIE_SOLITARY_EVAL_INTERVAL_FRAMES",
    "ZOMBIE_DOG_PATROL_SPEED",
    "ZOMBIE_DOG_ASSAULT_SPEED",
    "ZOMBIE_DOG_SIGHT_RANGE",
    "ZOMBIE_DOG_SURVIVOR_SIGHT_RANGE",
    "ZOMBIE_DOG_LONG_AXIS_RATIO",
    "ZOMBIE_DOG_SHORT_AXIS_RATIO",
    "ZOMBIE_DOG_WANDER_INTERVAL_MS",
    "ZOMBIE_DOG_PACK_CHASE_RANGE",
    "ZOMBIE_DOG_CHARGE_OFFSET_MIN_NORMAL",
    "ZOMBIE_DOG_CHARGE_OFFSET_MAX_NORMAL",
    "ZOMBIE_DOG_CHARGE_OFFSET_MIN_NIMBLE",
    "ZOMBIE_DOG_CHARGE_OFFSET_MAX_NIMBLE",
    "ZOMBIE_DOG_CHARGE_DISTANCE_NORMAL",
    "ZOMBIE_DOG_CHARGE_DISTANCE_NIMBLE",
    "ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NORMAL",
    "ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NIMBLE",
    "MovingFloorDirection",
    "MOVING_FLOOR_SPEED",
    "PATROL_BOT_SPRITE_SIZE",
    "PATROL_BOT_COLLISION_RADIUS",
    "PATROL_BOT_DIRECTION_COMMAND_RADIUS",
    "PATROL_BOT_DIRECTION_COMMAND_HOLD_MS",
    "PATROL_BOT_SPEED",
    "PATROL_BOT_HUMANOID_PAUSE_MS",
    "PATROL_BOT_ZOMBIE_DAMAGE_INTERVAL_FRAMES",
    "PATROL_BOT_PARALYZE_MS",
    "PATROL_BOT_PARALYZE_BLINK_MS",
    "PATROL_BOT_PARALYZE_MARKER_COLOR",
    "PATROL_BOT_COLLISION_MARGIN",
    "CAR_WIDTH",
    "CAR_HEIGHT",
    "CAR_SPEED",
    "CAR_HEALTH",
    "CAR_WALL_DAMAGE",
    "CAR_PITFALL_FALL_CENTER_RATIO",
    "FUEL_CAN_WIDTH",
    "FUEL_CAN_HEIGHT",
    "EMPTY_FUEL_CAN_WIDTH",
    "EMPTY_FUEL_CAN_HEIGHT",
    "FUEL_STATION_WIDTH",
    "FUEL_STATION_HEIGHT",
    "SPIKY_PLANT_HEALTH",
    "SPIKY_PLANT_RADIUS",
    "SPIKY_PLANT_COLLISION_RADIUS",
    "SPIKY_PLANT_HUMANOID_SPEED_FACTOR",
    "SPIKY_PLANT_CAR_SPEED_FACTOR",
    "PUDDLE_SPEED_FACTOR",
    "TRAPPED_ZOMBIE_SLOW_FACTOR",
    "TRAPPED_ZOMBIE_REPEL_PER_STACK",
    "TRAPPED_ZOMBIE_REPEL_MAX_MULT",
    "TRAPPED_ZOMBIE_REPEL_RADIUS_CELLS",
    "INTERNAL_WALL_HEALTH",
    "INTERNAL_WALL_BEVEL_DEPTH",
    "STEEL_BEAM_HEALTH",
    "PLAYER_WALL_DAMAGE",
    "BUDDY_WALL_DAMAGE",
    "ZOMBIE_WALL_DAMAGE",
    "PATROL_BOT_ZOMBIE_DAMAGE",
]



from __future__ import annotations

import pygame

from ..colors import (
    DAWN_AMBIENT_PALETTE_KEY,
    WHITE,
    ambient_palette_key_for_flashlights,
)
from ..entities_constants import SURVIVOR_MAX_SAFE_PASSENGERS
from ..localization import translate as tr
from ..models import (
    FuelMode,
    FuelProgress,
    GameClock,
    GameData,
    Groups,
    LevelLayout,
    ProgressState,
    Stage,
    TimedMessage,
)
from ..screen_constants import FPS
from ..entities import Camera
from .ambient import _set_ambient_palette
from .constants import INTRO_MESSAGE_DISPLAY_FRAMES
from ..render.decay_effects import prepare_decay_mask
from .spatial_index import SPATIAL_INDEX_CELL_SIZE, SpatialIndex


def frames_to_ms(frames: int) -> int:
    if frames <= 0:
        return 0
    return max(1, int(round((1000 / max(1, FPS)) * frames)))


def ms_to_frames(ms: int) -> int:
    if ms <= 0:
        return 0
    return max(1, int(round((max(1, FPS) / 1000) * ms)))


def schedule_timed_message(
    state: ProgressState,
    text: str | None,
    *,
    duration_frames: int,
    clear_on_input: bool = False,
    color: tuple[int, int, int] | None = None,
    align: str = "center",
    now_ms: int,
) -> None:
    if not text:
        state.timed_message = None
        return
    duration_ms = frames_to_ms(duration_frames)
    state.timed_message = TimedMessage(
        text=text,
        expires_at_ms=now_ms + duration_ms,
        clear_on_input=clear_on_input,
        color=color,
        align=align,
    )


def initialize_game_state(stage: Stage) -> GameData:
    """Initialize and return the base game state objects."""
    fuel_progress = (
        FuelProgress.FULL_CAN
        if stage.fuel_mode == FuelMode.START_FULL
        else FuelProgress.NONE
    )
    if stage.endurance_stage:
        fuel_progress = FuelProgress.NONE
    starts_with_flashlight = False
    initial_flashlights = 1 if starts_with_flashlight else 0
    initial_palette_key = ambient_palette_key_for_flashlights(initial_flashlights)
    intro_message = tr(stage.intro_key) if stage.intro_key else None
    game_state = ProgressState(
        game_over=False,
        game_won=False,
        timed_message=None,
        fade_in_started_at_ms=None,
        game_over_at=None,
        scaled_overview=None,
        overview_created=False,
        footprints=[],
        spatial_index=SpatialIndex(cell_size=SPATIAL_INDEX_CELL_SIZE),
        decay_effects=[],
        last_footprint_pos=None,
        footprint_visible_toggle=True,
        clock=GameClock(),
        fuel_progress=fuel_progress,
        flashlight_count=initial_flashlights,
        shoes_count=0,
        ambient_palette_key=initial_palette_key,
        hint_expires_at=0,
        hint_target_type=None,
        contact_hint_records=[],
        buddy_rescued=0,
        buddy_onboard=0,
        buddy_merged_count=0,
        survivors_onboard=0,
        survivors_rescued=0,
        survivor_messages=[],
        survivor_capacity=SURVIVOR_MAX_SAFE_PASSENGERS,
        seed=None,
        endurance_elapsed_ms=0,
        endurance_goal_ms=max(0, stage.endurance_goal_ms),
        dawn_ready=False,
        dawn_prompt_at=None,
        time_accel_active=False,
        time_accel_multiplier=1.0,
        last_zombie_spawn_time=0,
        dawn_carbonized=False,
        debug_mode=False,
        show_fps=False,
        falling_zombies=[],
        falling_spawn_carry=0,
        dust_rings=[],
        electrified_cells=set(),
        player_wall_target_cell=None,
        player_wall_target_ttl=0,
    )
    if intro_message:
        schedule_timed_message(
            game_state,
            intro_message,
            duration_frames=INTRO_MESSAGE_DISPLAY_FRAMES,
            clear_on_input=True,
            color=WHITE,
            align="left",
            now_ms=game_state.clock.elapsed_ms,
        )

    # Start fade-in from black when gameplay begins.
    game_state.fade_in_started_at_ms = game_state.clock.elapsed_ms

    prepare_decay_mask()

    # Create sprite groups
    all_sprites = pygame.sprite.LayeredUpdates()
    wall_group = pygame.sprite.Group()
    zombie_group = pygame.sprite.Group()
    survivor_group = pygame.sprite.Group()
    patrol_bot_group = pygame.sprite.Group()

    # Create camera
    cell_size = stage.cell_size
    level_width = stage.grid_cols * cell_size
    level_height = stage.grid_rows * cell_size
    camera = Camera(level_width, level_height)

    # Define level layout (will be filled by blueprint generation)
    field_rect = pygame.Rect(0, 0, level_width, level_height)

    return GameData(
        state=game_state,
        groups=Groups(
            all_sprites=all_sprites,
            wall_group=wall_group,
            zombie_group=zombie_group,
            survivor_group=survivor_group,
            patrol_bot_group=patrol_bot_group,
        ),
        camera=camera,
        layout=LevelLayout(
            field_rect=field_rect,
            grid_cols=stage.grid_cols,
            grid_rows=stage.grid_rows,
            outside_cells=set(),
            walkable_cells=[],
            outer_wall_cells=set(),
            wall_cells=set(),
            steel_beam_cells=set(),
            pitfall_cells=set(),
            car_walkable_cells=set(),
            car_spawn_cells=[],
            fall_spawn_cells=set(),
            spiky_plant_cells=set(),
            fire_floor_cells=set(),
            metal_floor_cells=set(),
            zombie_contaminated_cells=set(),
            puddle_cells=set(),
            bevel_corners={},
            moving_floor_cells={},
        ),
        fog={
            "hatch_patterns": {},
            "overlays": {},
        },
        stage=stage,
        cell_size=cell_size,
        wall_index=None,
        wall_index_dirty=True,
        blueprint=None,
        fuel=None,
        empty_fuel_can=None,
        fuel_station=None,
        flashlights=[],
        shoes=[],
    )


def carbonize_outdoor_zombies(game_data: GameData) -> None:
    """Petrify zombies that have already broken through to the exterior."""
    outside_cells = game_data.layout.outside_cells
    if not outside_cells:
        return
    cell_size = game_data.cell_size
    if cell_size <= 0:
        return
    group = game_data.groups.zombie_group
    if not group:
        return
    for zombie in list(group):
        if not zombie.alive():
            continue
        cell = (
            int(zombie.rect.centerx // cell_size),
            int(zombie.rect.centery // cell_size),
        )
        if cell in outside_cells:
            zombie.carbonize()


def update_endurance_timer(game_data: GameData, dt_ms: int) -> None:
    """Advance the endurance countdown and trigger dawn handoff."""
    stage = game_data.stage
    state = game_data.state
    if not stage.endurance_stage:
        return
    if state.endurance_goal_ms <= 0 or dt_ms <= 0:
        return
    state.endurance_elapsed_ms = min(
        state.endurance_goal_ms,
        state.endurance_elapsed_ms + dt_ms,
    )
    if not state.dawn_ready and state.endurance_elapsed_ms >= state.endurance_goal_ms:
        state.dawn_ready = True
        clock = getattr(state, "clock", None)
        state.dawn_prompt_at = clock.elapsed_ms if clock is not None else 0
        _set_ambient_palette(game_data, DAWN_AMBIENT_PALETTE_KEY, force=True)
    if state.dawn_ready:
        carbonize_outdoor_zombies(game_data)
        state.dawn_carbonized = True



from __future__ import annotations

from typing import Callable

from .patrol_paralyze import update_paralyze_from_patrol_contact


class ZombieVitals:
    def __init__(
        self,
        *,
        max_health: int,
        decay_duration_frames: float,
        decay_min_speed_ratio: float,
        carbonize_decay_frames: float,
        on_health_ratio: Callable[[float], None],
        on_kill: Callable[[], None],
        on_carbonize: Callable[[], None],
    ) -> None:
        self.max_health = max(1, int(max_health))
        self.health = self.max_health
        self.decay_duration_frames = float(decay_duration_frames)
        self.decay_min_speed_ratio = float(decay_min_speed_ratio)
        self.carbonize_decay_frames = float(carbonize_decay_frames)
        self.on_health_ratio = on_health_ratio
        self.on_kill = on_kill
        self.on_carbonize = on_carbonize
        self.carbonized = False
        self.decay_carry = 0.0
        self.last_damage_ms: int | None = None
        self.last_damage_source: str | None = None
        self.patrol_paralyze_until_ms = 0
        self.patrol_damage_frame_counter = 0

    def _set_health(self, new_health: int) -> None:
        self.health = max(0, min(self.max_health, int(new_health)))
        health_ratio = 0.0 if self.max_health <= 0 else self.health / self.max_health
        health_ratio = max(0.0, min(1.0, health_ratio))
        speed_ratio = (
            self.decay_min_speed_ratio
            + (1.0 - self.decay_min_speed_ratio) * health_ratio
        )
        self.on_health_ratio(speed_ratio)
        if self.health <= 0:
            self.on_kill()

    def apply_decay(self) -> None:
        if self.decay_duration_frames <= 0:
            return
        self.decay_carry += self.max_health / self.decay_duration_frames
        if self.decay_carry >= 1.0:
            decay_amount = int(self.decay_carry)
            self.decay_carry -= decay_amount
            self._set_health(self.health - decay_amount)

    def take_damage(
        self, amount: int, *, source: str | None = None, now_ms: int
    ) -> None:
        if amount <= 0:
            return
        self.last_damage_ms = now_ms
        self.last_damage_source = source
        self._set_health(self.health - amount)

    def carbonize(self) -> None:
        if self.carbonized:
            return
        self.carbonized = True
        if self.decay_duration_frames > 0:
            remaining_ratio = min(
                1.0, self.carbonize_decay_frames / self.decay_duration_frames
            )
            remaining_health = max(1, int(round(self.max_health * remaining_ratio)))
            self.health = min(self.health, remaining_health)
            self.decay_carry = 0.0
        self.on_health_ratio(0.0)
        self.on_carbonize()

    def update_patrol_paralyze(
        self,
        *,
        entity_center: tuple[float, float],
        entity_radius: float,
        patrol_bots: list,
        now_ms: int,
        paralyze_duration_ms: int,
        damage_interval_frames: int,
        damage_amount: int,
        apply_damage: Callable[[int], None],
    ) -> bool:
        _, self.patrol_paralyze_until_ms, self.patrol_damage_frame_counter = (
            update_paralyze_from_patrol_contact(
                entity_center=entity_center,
                entity_radius=entity_radius,
                patrol_bots=patrol_bots,
                now_ms=now_ms,
                paralyze_until_ms=self.patrol_paralyze_until_ms,
                paralyze_duration_ms=paralyze_duration_ms,
                damage_counter=self.patrol_damage_frame_counter,
                damage_interval_frames=damage_interval_frames,
                damage_amount=damage_amount,
                apply_damage=apply_damage,
            )
        )
        return now_ms < self.patrol_paralyze_until_ms

    def update_patrol_floor_paralyze(
        self,
        *,
        on_electrified_floor: bool,
        now_ms: int,
        paralyze_duration_ms: int,
        damage_interval_frames: int,
        damage_amount: int,
        apply_damage: Callable[[int], None] | None,
    ) -> bool:
        if on_electrified_floor:
            self.patrol_paralyze_until_ms = max(
                self.patrol_paralyze_until_ms,
                now_ms + paralyze_duration_ms,
            )
            if (
                damage_interval_frames > 0
                and apply_damage is not None
                and damage_amount > 0
            ):
                self.patrol_damage_frame_counter = (
                    self.patrol_damage_frame_counter + 1
                ) % damage_interval_frames
                if self.patrol_damage_frame_counter == 0:
                    apply_damage(damage_amount)
        return now_ms < self.patrol_paralyze_until_ms



"""Window and presentation helpers for zombie_escape."""

from __future__ import annotations

from typing import TYPE_CHECKING
import os

import pygame
from pygame import surface

from .screen_constants import (
    DEFAULT_WINDOW_SCALE,
    SCREEN_HEIGHT,
    SCREEN_WIDTH,
    WINDOW_SCALE_MAX,
    WINDOW_SCALE_MIN,
)

if TYPE_CHECKING:  # pragma: no cover - typing only
    from .models import GameData

current_window_scale = DEFAULT_WINDOW_SCALE  # Applied to the OS window only
current_maximized = False
last_window_scale = DEFAULT_WINDOW_SCALE
last_window_position: tuple[int, int] | None = None
current_window_size = (
    int(SCREEN_WIDTH * DEFAULT_WINDOW_SCALE),
    int(SCREEN_HEIGHT * DEFAULT_WINDOW_SCALE),
)
last_logged_window_size = current_window_size
_scaled_logical_size = (SCREEN_WIDTH, SCREEN_HEIGHT)
_WINDOW_SCALE_COOLDOWN_MS = 500
_last_window_scale_apply_ms: int | None = None
_pending_window_scale: float | None = None
_pending_window_scale_game_data: "GameData | None" = None
_VSYNC_ENV = os.environ.get("ZOMBIE_ESCAPE_VSYNC")
_VSYNC_PREFERENCE: int | None
if _VSYNC_ENV is None:
    _VSYNC_PREFERENCE = None
else:
    _VSYNC_PREFERENCE = 1 if _VSYNC_ENV not in ("0", "false", "False") else 0

__all__ = [
    "present",
    "apply_window_scale",
    "prime_scaled_logical_size",
    "nudge_window_scale",
    "nudge_menu_window_scale",
    "toggle_fullscreen",
    "sync_window_size",
    "adjust_menu_logical_size",
    "set_scaled_logical_size",
]


def present(logical_surface: surface.Surface) -> None:
    """Scale the logical surface directly to the window and flip buffers."""
    _maybe_apply_pending_window_scale()
    window = pygame.display.get_surface()
    if window is None:
        return
    window_size = _fetch_window_size(window)
    _update_window_size(window_size, source="frame")
    logical_size = logical_surface.get_size()
    if _use_scaled_display():
        target_size = window.get_size()
        if logical_size == target_size:
            scaled_surface = logical_surface
        elif (
            logical_size[0] * 2 == target_size[0]
            and logical_size[1] * 2 == target_size[1]
        ):
            scaled_surface = pygame.transform.scale2x(logical_surface)
        else:
            scaled_surface = pygame.transform.scale(logical_surface, target_size)
        window.blit(scaled_surface, (0, 0))
    elif window_size == logical_size:
        window.blit(logical_surface, (0, 0))
    else:
        # Preserve aspect ratio with letterboxing.
        scale_x = window_size[0] / max(1, logical_size[0])
        scale_y = window_size[1] / max(1, logical_size[1])
        scale = min(scale_x, scale_y)
        scaled_width = max(1, int(logical_size[0] * scale))
        scaled_height = max(1, int(logical_size[1] * scale))
        window.fill((0, 0, 0))
        if (scaled_width, scaled_height) == logical_size:
            scaled_surface = logical_surface
        else:
            scaled_surface = pygame.transform.scale(
                logical_surface, (scaled_width, scaled_height)
            )
        offset_x = (window_size[0] - scaled_width) // 2
        offset_y = (window_size[1] - scaled_height) // 2
        window.blit(scaled_surface, (offset_x, offset_y))
    pygame.display.flip()


def apply_window_scale(
    scale: float, *, game_data: "GameData | None" = None
) -> surface.Surface:
    """Resize the OS window; logical render surface stays constant."""
    global \
        _last_window_scale_apply_ms, \
        _pending_window_scale, \
        _pending_window_scale_game_data

    now = pygame.time.get_ticks()
    if pygame.display.get_surface() is None:
        _last_window_scale_apply_ms = now
        _pending_window_scale = None
        _pending_window_scale_game_data = None
        return _apply_window_scale_now(scale, game_data=game_data)
    if (
        _last_window_scale_apply_ms is not None
        and now - _last_window_scale_apply_ms < _WINDOW_SCALE_COOLDOWN_MS
    ):
        _pending_window_scale = scale
        _pending_window_scale_game_data = game_data
        window = pygame.display.get_surface()
        return window if window is not None else pygame.Surface((1, 1))
    _last_window_scale_apply_ms = now
    _pending_window_scale = None
    _pending_window_scale_game_data = None
    return _apply_window_scale_now(scale, game_data=game_data)


def _apply_window_scale_now(
    scale: float, *, game_data: "GameData | None" = None
) -> surface.Surface:
    global current_window_scale, current_maximized, last_window_scale

    clamped_scale = max(WINDOW_SCALE_MIN, min(WINDOW_SCALE_MAX, scale))
    current_window_scale = clamped_scale
    last_window_scale = clamped_scale
    current_maximized = False

    window_width = max(1, int(SCREEN_WIDTH * current_window_scale))
    window_height = max(1, int(SCREEN_HEIGHT * current_window_scale))

    flags = pygame.RESIZABLE
    if _use_scaled_display():
        flags |= pygame.SCALED
        new_window = _set_mode(_scaled_logical_size, flags)
        _set_window_size((window_width, window_height))
    else:
        new_window = _set_mode((window_width, window_height), flags)
    _update_window_size((window_width, window_height), source="apply_scale")
    _update_window_caption()

    if game_data is not None:
        game_data.state.overview_created = False

    return new_window


def prime_scaled_logical_size(size: tuple[int, int]) -> None:
    """Set initial logical render size before the first window is created."""
    global _scaled_logical_size
    target = _normalize_window_size(size)
    _scaled_logical_size = target


def nudge_window_scale(
    multiplier: float, *, game_data: "GameData | None" = None
) -> surface.Surface:
    """Scale the window relative to the current zoom level."""
    delta = 1.0 if multiplier >= 1.0 else -1.0
    target_scale = current_window_scale + delta
    return apply_window_scale(target_scale, game_data=game_data)


def nudge_menu_window_scale(
    multiplier: float, *, game_data: "GameData | None" = None
) -> surface.Surface:
    """Scale the window and update menu logical size consistently."""
    delta = 1.0 if multiplier >= 1.0 else -1.0
    target_scale = current_window_scale + delta
    set_scaled_logical_size(
        (SCREEN_WIDTH, SCREEN_HEIGHT), preserve_window_size=False, game_data=game_data
    )
    return apply_window_scale(target_scale, game_data=game_data)


def toggle_fullscreen(*, game_data: "GameData | None" = None) -> surface.Surface | None:
    """Toggle fullscreen without persisting the setting."""
    global current_maximized, last_window_scale, last_window_position
    if current_maximized:
        current_maximized = False
        window_width = max(1, int(SCREEN_WIDTH * last_window_scale))
        window_height = max(1, int(SCREEN_HEIGHT * last_window_scale))
        _set_sdl2_fullscreen(False, None)
        flags = pygame.RESIZABLE
        if _use_scaled_display():
            flags |= pygame.SCALED
            window = _set_mode(_scaled_logical_size, flags)
            _set_window_size((window_width, window_height))
        else:
            window = _set_mode((window_width, window_height), flags)
        if last_window_position is not None:
            _restore_window_position(last_window_position)
        _restore_window()
        _update_window_caption()
        _update_window_size((window_width, window_height), source="toggle_windowed")
    else:
        last_window_scale = current_window_scale
        last_window_position = _fetch_window_position()
        current_maximized = True
        display_index = _fetch_display_index()
        window = None
        if _set_sdl2_fullscreen(True, display_index):
            window = pygame.display.get_surface()
        if window is None:
            flags = pygame.FULLSCREEN
            render_size = (0, 0)
            if _use_scaled_display():
                flags |= pygame.SCALED
                render_size = _scaled_logical_size
            if display_index is None:
                window = _set_mode(render_size, flags)
            else:
                window = _set_mode(render_size, flags, display=display_index)
        window_width, window_height = _fetch_window_size(window)
        _update_window_caption()
        _update_window_size((window_width, window_height), source="toggle_fullscreen")
    if game_data is not None:
        game_data.state.overview_created = False
    return window


def sync_window_size(
    event: pygame.event.Event, *, game_data: "GameData | None" = None
) -> None:
    """Synchronize tracked window size with SDL window events."""
    global current_window_scale, last_window_scale, _last_window_scale_apply_ms
    size = getattr(event, "size", None)
    if not size:
        width = getattr(event, "x", None)
        height = getattr(event, "y", None)
        if width is not None and height is not None:
            size = (width, height)
    if not size:
        return
    window_width, window_height = _normalize_window_size(size)
    _update_window_size((window_width, window_height), source="window_event")
    if not current_maximized:
        scale_x = window_width / max(1, SCREEN_WIDTH)
        scale_y = window_height / max(1, SCREEN_HEIGHT)
        scale = max(WINDOW_SCALE_MIN, min(WINDOW_SCALE_MAX, min(scale_x, scale_y)))
        current_window_scale = scale
        last_window_scale = scale
    _update_window_caption()
    if game_data is not None:
        game_data.state.overview_created = False
    _last_window_scale_apply_ms = None
    _maybe_apply_pending_window_scale()


def set_scaled_logical_size(
    size: tuple[int, int],
    *,
    preserve_window_size: bool = True,
    game_data: "GameData | None" = None,
) -> None:
    """Update the logical render size when using pygame.SCALED."""
    global _scaled_logical_size
    if not _use_scaled_display():
        return
    target = _normalize_window_size(size)
    if target == _scaled_logical_size:
        return
    previous_window_size = _fetch_window_size(pygame.display.get_surface())
    _scaled_logical_size = target
    flags = pygame.SCALED
    if current_maximized:
        flags |= pygame.FULLSCREEN
        _set_mode(_scaled_logical_size, flags)
    else:
        flags |= pygame.RESIZABLE
        _set_mode(_scaled_logical_size, flags)
        if preserve_window_size:
            _set_window_size(previous_window_size)
    _update_window_caption()
    if game_data is not None:
        game_data.state.overview_created = False


def adjust_menu_logical_size(*, game_data: "GameData | None" = None) -> None:
    """Match menu render size to the current window scale."""
    set_scaled_logical_size((SCREEN_WIDTH, SCREEN_HEIGHT), game_data=game_data)


def _fetch_window_size(window: surface.Surface | None) -> tuple[int, int]:
    if hasattr(pygame.display, "get_window_size"):
        size = pygame.display.get_window_size()
        if size != (0, 0):
            return _normalize_window_size(size)
    if window is not None:
        return _normalize_window_size(window.get_size())
    window_width = max(1, int(SCREEN_WIDTH * last_window_scale))
    window_height = max(1, int(SCREEN_HEIGHT * last_window_scale))
    return window_width, window_height


def _normalize_window_size(size: tuple[int, int]) -> tuple[int, int]:
    width = max(1, int(size[0]))
    height = max(1, int(size[1]))
    return width, height


def _set_window_size(size: tuple[int, int]) -> None:
    setter = getattr(pygame.display, "set_window_size", None)
    if callable(setter):
        try:
            setter(size)
            return
        except Exception:
            pass
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return
    setter = getattr(window, "set_size", None)
    if callable(setter):
        try:
            setter(size)
            return
        except Exception:
            return
    try:
        window.size = size
    except Exception:
        return


def _set_mode(
    size: tuple[int, int], flags: int, *, display: int | None = None
) -> surface.Surface:
    vsync = _VSYNC_PREFERENCE
    try:
        kwargs: dict[str, int] = {}
        if display is not None:
            kwargs["display"] = int(display)
        if vsync is None:
            return pygame.display.set_mode(size, flags, **kwargs)
        kwargs["vsync"] = int(vsync)
        return pygame.display.set_mode(size, flags, **kwargs)
    except TypeError:
        if display is None:
            return pygame.display.set_mode(size, flags)
        return pygame.display.set_mode(size, flags, display=int(display))


def _use_scaled_display() -> bool:
    return hasattr(pygame, "SCALED")


def _fetch_window_position() -> tuple[int, int] | None:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return None
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return None
    try:
        position = window.position
    except Exception:
        return None
    if not position:
        return None
    return (int(position[0]), int(position[1]))


def _restore_window_position(position: tuple[int, int]) -> None:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return
    setter = getattr(window, "set_position", None)
    if setter is not None:
        try:
            setter(position)
            return
        except Exception:
            return
    try:
        window.position = position
    except Exception:
        return


def _fetch_display_index() -> int | None:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return None
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return None

    display_index = _infer_display_index_from_position(window, sdl2)
    if display_index is not None:
        return display_index

    try:
        return window.get_display_index()
    except Exception:
        return None


def _infer_display_index_from_position(window: object, sdl2: object) -> int | None:
    try:
        position = window.position  # type: ignore[attr-defined]
    except Exception:
        return None
    if not position:
        return None

    center_x, center_y = _window_center_from_position(window, position)
    display_count = _get_display_count(sdl2)
    if display_count is None:
        return None

    for display_index in range(display_count):
        bounds = _get_display_bounds(sdl2, display_index)
        if bounds is None:
            continue
        x, y, width, height = bounds
        if x <= center_x < x + width and y <= center_y < y + height:
            return display_index
    return None


def _window_center_from_position(
    window: object, position: tuple[int, int]
) -> tuple[int, int]:
    x, y = position
    try:
        width, height = window.size  # type: ignore[attr-defined]
    except Exception:
        width, height = _fetch_window_size(None)
    return x + width // 2, y + height // 2


def _get_display_count(sdl2: object) -> int | None:
    candidate = getattr(sdl2, "get_num_video_displays", None)
    if candidate is None:
        candidate = getattr(
            getattr(sdl2, "video", None), "get_num_video_displays", None
        )
    if candidate is None:
        return None
    try:
        return int(candidate())
    except Exception:
        return None


def _get_display_bounds(
    sdl2: object, display_index: int
) -> tuple[int, int, int, int] | None:
    candidate = getattr(sdl2, "get_display_bounds", None)
    if candidate is None:
        candidate = getattr(getattr(sdl2, "video", None), "get_display_bounds", None)
    if candidate is None:
        return None
    try:
        bounds = candidate(display_index)
    except Exception:
        return None
    if bounds is None:
        return None
    if hasattr(bounds, "x"):
        return (int(bounds.x), int(bounds.y), int(bounds.w), int(bounds.h))
    if isinstance(bounds, (tuple, list)) and len(bounds) >= 4:
        return (int(bounds[0]), int(bounds[1]), int(bounds[2]), int(bounds[3]))
    return None


def _set_sdl2_fullscreen(enable: bool, display_index: int | None) -> bool:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return False
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return False

    if enable and display_index is not None:
        setter = getattr(window, "set_display_index", None)
        if setter is not None:
            try:
                setter(display_index)
            except Exception:
                pass

    if hasattr(window, "fullscreen"):
        try:
            window.fullscreen = enable
            return True
        except Exception:
            pass

    setter = getattr(window, "set_fullscreen", None)
    if setter is None:
        return False
    try:
        setter(enable)
        return True
    except Exception:
        pass

    if enable:
        for attr_name in (
            "WINDOW_FULLSCREEN_DESKTOP",
            "FULLSCREEN_DESKTOP",
            "WINDOW_FULLSCREEN",
        ):
            mode = getattr(sdl2, attr_name, None)
            if mode is None:
                continue
            try:
                setter(mode)
                return True
            except Exception:
                continue
    return False


def _update_window_size(size: tuple[int, int], *, source: str) -> None:
    global current_window_size, last_logged_window_size
    _ = source
    current_window_size = size
    if size != last_logged_window_size:
        print(f"WINDOW_SIZE {size[0]}x{size[1]}")
        last_logged_window_size = size


def _update_window_caption() -> None:
    pygame.display.set_caption("Zombie Escape")


def _restore_window() -> None:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return
    try:
        window.restore()
    except Exception:
        return


def _maybe_apply_pending_window_scale() -> None:
    global \
        _last_window_scale_apply_ms, \
        _pending_window_scale, \
        _pending_window_scale_game_data
    if _pending_window_scale is None or current_maximized:
        return
    now = pygame.time.get_ticks()
    if (
        _last_window_scale_apply_ms is not None
        and now - _last_window_scale_apply_ms < _WINDOW_SCALE_COOLDOWN_MS
    ):
        return
    scale = _pending_window_scale
    game_data = _pending_window_scale_game_data
    _pending_window_scale = None
    _pending_window_scale_game_data = None
    _last_window_scale_apply_ms = now
    _apply_window_scale_now(scale, game_data=game_data)



from __future__ import annotations

import math
from typing import Any, Sequence

import pygame

from ..entities import (
    Car,
    Player,
    PatrolBot,
    Survivor,
    Wall,
    Zombie,
    ZombieDog,
)
from ..entities_constants import (
    SPIKY_PLANT_HUMANOID_SPEED_FACTOR,
    SPIKY_PLANT_CAR_SPEED_FACTOR,
    PUDDLE_SPEED_FACTOR,
    HUMANOID_WALL_BUMP_FRAMES,
    MOVING_FLOOR_SPEED,
    PLAYER_SPEED,
    ZombieKind,
    ZOMBIE_DOG_PACK_CHASE_RANGE,
    ZOMBIE_DOG_SURVIVOR_SIGHT_RANGE,
    ZOMBIE_LINEFORMER_JOIN_RADIUS,
    ZOMBIE_SEPARATION_DISTANCE,
    ZOMBIE_TRACKER_CROWD_BAND_WIDTH,
    ZOMBIE_TRACKER_GRID_CROWD_COUNT,
    ZOMBIE_WALL_HUG_SENSOR_DISTANCE,
)
from ..gameplay_constants import (
    SHOES_SPEED_MULTIPLIER_ONE,
    SHOES_SPEED_MULTIPLIER_TWO,
)
from ..models import FallingEntity, GameData
from ..rng import get_rng
from ..surface_effects import resolve_surface_speed_factor
from ..entities.movement_helpers import pitfall_target
from ..world_grid import WallIndex, apply_cell_edge_nudge, walls_for_radius
from .moving_floor import get_floor_overlap_rect, get_moving_floor_drift
from .constants import LAYER_PLAYERS, MAX_ZOMBIES
from ..render.decay_effects import DecayingEntityEffect, update_decay_effects
from .spawn import spawn_weighted_zombie, update_falling_zombies
from .spatial_index import SpatialKind
from .survivors import update_survivors
from .utils import (
    find_nearby_offscreen_spawn_position,
    is_entity_in_fov,
    rect_visible_on_screen,
)

RNG = get_rng()


def _cell_at_position(
    x: float,
    y: float,
    *,
    cell_size: int,
) -> tuple[int, int] | None:
    if cell_size <= 0:
        return None
    return (int(x // cell_size), int(y // cell_size))


def _is_on_fire_floor(
    x: float,
    y: float,
    *,
    cell_size: int,
    fire_floor_cells: set[tuple[int, int]],
) -> bool:
    cell = _cell_at_position(x, y, cell_size=cell_size)
    return bool(cell is not None and cell in fire_floor_cells)


def process_player_input(
    keys: Sequence[bool],
    player: Player,
    car: Car | None,
    shoes_count: int = 0,
    pad_input: tuple[float, float] = (0.0, 0.0),
) -> tuple[float, float, float, float]:
    """Process keyboard input and return movement deltas."""
    dx_input, dy_input = 0, 0
    if keys[pygame.K_w] or keys[pygame.K_UP]:
        dy_input -= 1
    if keys[pygame.K_s] or keys[pygame.K_DOWN]:
        dy_input += 1
    if keys[pygame.K_a] or keys[pygame.K_LEFT]:
        dx_input -= 1
    if keys[pygame.K_d] or keys[pygame.K_RIGHT]:
        dx_input += 1
    dx_input += pad_input[0]
    dy_input += pad_input[1]

    player.update_facing_from_input(dx_input, dy_input)

    player_dx, player_dy, car_dx, car_dy = 0, 0, 0, 0

    if player.in_car and car and car.alive():
        car.update_facing_from_input(dx_input, dy_input)
        target_speed = car.speed
        move_len = math.hypot(dx_input, dy_input)
        if move_len > 0:
            car_dx, car_dy = (
                (dx_input / move_len) * target_speed,
                (dy_input / move_len) * target_speed,
            )
    elif not player.in_car:
        target_speed = PLAYER_SPEED * _shoes_speed_multiplier(shoes_count)
        move_len = math.hypot(dx_input, dy_input)
        if move_len > 0:
            player_dx, player_dy = (
                (dx_input / move_len) * target_speed,
                (dy_input / move_len) * target_speed,
            )

    return player_dx, player_dy, car_dx, car_dy


def _shoes_speed_multiplier(shoes_count: int) -> float:
    count = max(0, int(shoes_count))
    if count >= 2:
        return SHOES_SPEED_MULTIPLIER_TWO
    if count == 1:
        return SHOES_SPEED_MULTIPLIER_ONE
    return 1.0


def update_entities(
    game_data: GameData,
    player_dx: float,
    player_dy: float,
    car_dx: float,
    car_dy: float,
    config: dict[str, Any],
    wall_index: WallIndex | None = None,
) -> None:
    """Update positions and states of game entities."""
    player = game_data.player
    assert player is not None
    car = game_data.car
    wall_group = game_data.groups.wall_group
    all_sprites = game_data.groups.all_sprites
    zombie_group = game_data.groups.zombie_group
    survivor_group = game_data.groups.survivor_group
    patrol_bot_group = game_data.groups.patrol_bot_group
    spatial_index = game_data.state.spatial_index
    camera = game_data.camera
    stage = game_data.stage
    active_car = car if car and car.alive() else None
    pitfall_cells = game_data.layout.pitfall_cells
    fire_floor_cells = game_data.layout.fire_floor_cells
    field_rect = game_data.layout.field_rect
    current_time = game_data.state.clock.elapsed_ms

    all_walls = list(wall_group) if wall_index is None else None

    def _walls_near(center: tuple[float, float], radius: float) -> list[Wall]:
        if wall_index is None:
            return all_walls or []
        return walls_for_radius(
            wall_index,
            center,
            radius,
            cell_size=game_data.cell_size,
            grid_cols=game_data.layout.grid_cols,
            grid_rows=game_data.layout.grid_rows,
        )

    # Update player/car movement
    if player.in_car and active_car:
        player.on_moving_floor = False
        floor_dx, floor_dy = get_moving_floor_drift(
            active_car.rect,
            game_data.layout,
            cell_size=game_data.cell_size,
            speed=MOVING_FLOOR_SPEED,
        )
        car_dx += floor_dx
        car_dy += floor_dy

        # Spiky plant / Puddle slow-down
        speed_factor = resolve_surface_speed_factor(
            active_car.x,
            active_car.y,
            active_car.collision_radius,
            cell_size=game_data.cell_size,
            puddle_cells=game_data.layout.puddle_cells,
            spiky_plants=game_data.spiky_plants,
            spiky_plant_speed_factor=SPIKY_PLANT_CAR_SPEED_FACTOR,
            puddle_speed_factor=PUDDLE_SPEED_FACTOR,
        )

        if speed_factor != 1.0:
            car_dx *= speed_factor
            car_dy *= speed_factor

        car_dx, car_dy = apply_cell_edge_nudge(
            active_car.x,
            active_car.y,
            car_dx,
            car_dy,
            layout=game_data.layout,
            cell_size=game_data.cell_size,
        )
        car_walls = _walls_near((active_car.x, active_car.y), 150.0)
        active_car.move(
            car_dx,
            car_dy,
            car_walls,
            walls_nearby=wall_index is not None,
            cell_size=game_data.cell_size,
            pitfall_cells=pitfall_cells,
        )
        if getattr(active_car, "pending_pitfall_fall", False):
            active_car.health = 0
        if field_rect is not None:
            car_allow_rect = field_rect.inflate(
                active_car.rect.width, active_car.rect.height
            )
            clamped_rect = active_car.rect.clamp(car_allow_rect)
            if clamped_rect.topleft != active_car.rect.topleft:
                active_car.rect = clamped_rect
                active_car.x = float(active_car.rect.centerx)
                active_car.y = float(active_car.rect.centery)
        player.rect.center = active_car.rect.center
        player.x, player.y = active_car.x, active_car.y
    elif not player.in_car:
        # Ensure player is in all_sprites if not in car
        if player not in all_sprites and not game_data.state.game_over:
            all_sprites.add(player, layer=LAYER_PLAYERS)
        player.pending_pitfall_fall = False
        movement_locked = game_data.state.game_over
        floor_dx = 0.0
        floor_dy = 0.0
        if not movement_locked:
            floor_dx, floor_dy = get_moving_floor_drift(
                get_floor_overlap_rect(player),
                game_data.layout,
                cell_size=game_data.cell_size,
                speed=MOVING_FLOOR_SPEED,
            )
            player_dx += floor_dx
            player_dy += floor_dy

            # Spiky plant / Puddle slow-down
            speed_factor = resolve_surface_speed_factor(
                player.x,
                player.y,
                player.collision_radius,
                cell_size=game_data.cell_size,
                puddle_cells=game_data.layout.puddle_cells,
                spiky_plants=game_data.spiky_plants,
                spiky_plant_speed_factor=SPIKY_PLANT_HUMANOID_SPEED_FACTOR,
                puddle_speed_factor=PUDDLE_SPEED_FACTOR,
            )

            if speed_factor != 1.0:
                player_dx *= speed_factor
                player_dy *= speed_factor
            player.on_moving_floor = abs(floor_dx) > 0.0 or abs(floor_dy) > 0.0
        else:
            player_dx = 0.0
            player_dy = 0.0
            player.on_moving_floor = False

        player_dx, player_dy = apply_cell_edge_nudge(
            player.x,
            player.y,
            player_dx,
            player_dy,
            layout=game_data.layout,
            cell_size=game_data.cell_size,
        )
        player.move(
            player_dx,
            player_dy,
            wall_group,
            patrol_bot_group=patrol_bot_group,
            wall_index=wall_index,
            cell_size=game_data.cell_size,
            layout=game_data.layout,
            now_ms=game_data.state.clock.elapsed_ms,
        )
        if (
            not game_data.state.game_over
            and not player.is_jumping
            and _is_on_fire_floor(
                player.x,
                player.y,
                cell_size=game_data.cell_size,
                fire_floor_cells=fire_floor_cells,
            )
        ):
            game_data.state.decay_effects.append(
                DecayingEntityEffect(player.image, player.rect.center)
            )
            if player in all_sprites:
                all_sprites.remove(player)
            game_data.state.game_over = True
            game_data.state.game_over_at = game_data.state.clock.elapsed_ms
            print("Player burned on fire floor!")
        if (
            getattr(player, "pending_pitfall_fall", False)
            and not game_data.state.game_over
        ):
            pull_dist = player.collision_radius * 0.5
            pitfall_target_pos = pitfall_target(
                x=player.x,
                y=player.y,
                cell_size=game_data.cell_size,
                pitfall_cells=pitfall_cells,
                pull_distance=pull_dist,
            )
            if pitfall_target_pos is None:
                pitfall_target_pos = player.rect.center
            game_data.state.falling_zombies.append(
                FallingEntity(
                    start_pos=(int(player.x), int(player.y)),
                    target_pos=pitfall_target_pos,
                    started_at_ms=game_data.state.clock.elapsed_ms,
                    pre_fx_ms=0,
                    fall_duration_ms=500,
                    dust_duration_ms=0,
                    kind=None,
                    mode="pitfall",
                )
            )
            if player in all_sprites:
                all_sprites.remove(player)
            game_data.state.game_over = True
            game_data.state.game_over_at = game_data.state.clock.elapsed_ms
            print("Player fell into pitfall!")
    else:
        # Player flagged as in-car but car is gone; drop them back to foot control
        player.in_car = False
        player.on_moving_floor = False

    # Update camera
    target_for_camera = active_car if player.in_car and active_car else player
    camera.update(target_for_camera)

    if player.inner_wall_hit and player.inner_wall_cell is not None:
        game_data.state.player_wall_target_cell = player.inner_wall_cell
        game_data.state.player_wall_target_ttl = HUMANOID_WALL_BUMP_FRAMES
    elif game_data.state.player_wall_target_ttl > 0:
        game_data.state.player_wall_target_ttl -= 1
        if game_data.state.player_wall_target_ttl <= 0:
            game_data.state.player_wall_target_cell = None

    wall_target_cell = (
        game_data.state.player_wall_target_cell
        if game_data.state.player_wall_target_ttl > 0
        else None
    )

    update_survivors(
        game_data,
        wall_index=wall_index,
        wall_target_cell=wall_target_cell,
        patrol_bot_group=patrol_bot_group,
    )
    for survivor in list(survivor_group):
        if not survivor.alive():
            continue
        if _is_on_fire_floor(
            survivor.x,
            survivor.y,
            cell_size=game_data.cell_size,
            fire_floor_cells=fire_floor_cells,
        ) and not getattr(survivor, "is_jumping", False):
            game_data.state.decay_effects.append(
                DecayingEntityEffect(survivor.image, survivor.rect.center)
            )
            survivor.kill()
            if getattr(survivor, "is_buddy", False):
                game_data.state.game_over = True
                if game_data.state.game_over_at is None:
                    game_data.state.game_over_at = game_data.state.clock.elapsed_ms
                print("Buddy burned on fire floor!")
            else:
                print("Survivor burned on fire floor!")
            continue
        if getattr(survivor, "pending_pitfall_fall", False):
            visible = rect_visible_on_screen(camera, survivor.rect)
            fov_target = active_car if player.in_car and active_car else player
            in_fov = is_entity_in_fov(
                survivor.rect,
                fov_target=fov_target,
                flashlight_count=game_data.state.flashlight_count,
            )
            if not (visible and in_fov):
                spawn_pos = find_nearby_offscreen_spawn_position(
                    game_data.layout.walkable_cells,
                    game_data.cell_size,
                    camera=camera,
                )
                survivor.teleport(spawn_pos)
                continue
            pull_dist = survivor.collision_radius * 0.5
            pitfall_target_pos = pitfall_target(
                x=survivor.x,
                y=survivor.y,
                cell_size=game_data.cell_size,
                pitfall_cells=pitfall_cells,
                pull_distance=pull_dist,
            )
            if pitfall_target_pos is None:
                pitfall_target_pos = survivor.rect.center
            game_data.state.falling_zombies.append(
                FallingEntity(
                    start_pos=(int(survivor.x), int(survivor.y)),
                    target_pos=pitfall_target_pos,
                    started_at_ms=game_data.state.clock.elapsed_ms,
                    pre_fx_ms=0,
                    fall_duration_ms=500,
                    dust_duration_ms=0,
                    kind=None,
                    mode="pitfall",
                )
            )
            survivor.kill()
            if getattr(survivor, "is_buddy", False):
                print("Buddy fell into pitfall!")
            else:
                print("Survivor fell into pitfall!")
    update_falling_zombies(game_data, config)

    # Spawn new zombies if needed
    spawn_interval = max(1, stage.spawn_interval_ms)
    spawn_blocked = stage.endurance_stage and game_data.state.dawn_ready
    if (
        len(zombie_group) < MAX_ZOMBIES
        and not spawn_blocked
        and current_time - game_data.state.last_zombie_spawn_time > spawn_interval
    ):
        spawn_count = max(1, int(stage.zombie_spawn_count_per_interval))
        spawned_any = False
        for _ in range(spawn_count):
            if len(zombie_group) >= MAX_ZOMBIES:
                break
            if spawn_weighted_zombie(game_data, config):
                spawned_any = True
        if spawned_any:
            game_data.state.last_zombie_spawn_time = current_time

    # Update zombies
    target_center = (
        active_car.rect.center if player.in_car and active_car else player.rect.center
    )
    buddies = [
        survivor
        for survivor in survivor_group
        if survivor.alive() and survivor.is_buddy and not survivor.rescued
    ]
    buddies_on_screen = [
        buddy for buddy in buddies if rect_visible_on_screen(camera, buddy.rect)
    ]

    survivors_on_screen: list[Survivor] = []
    if stage.survivor_rescue_stage:
        for survivor in survivor_group:
            if survivor.alive():
                if rect_visible_on_screen(camera, survivor.rect):
                    survivors_on_screen.append(survivor)

    game_data.lineformer_trains.pre_update(
        game_data,
        config=config,
        now_ms=current_time,
    )
    trapped_spiky_plant_counts: dict[tuple[int, int], int] = {}
    cell_size = game_data.cell_size
    spiky_plants = game_data.spiky_plants
    if cell_size > 0 and spiky_plants:
        for zombie in zombie_group:
            if not zombie.alive() or not getattr(zombie, "is_trapped", False):
                continue
            cell = (int(zombie.x // cell_size), int(zombie.y // cell_size))
            hp = spiky_plants.get(cell)
            if hp and hp.alive():
                trapped_spiky_plant_counts[cell] = (
                    trapped_spiky_plant_counts.get(cell, 0) + 1
                )
    zombies_sorted: list[Zombie | ZombieDog] = sorted(
        list(zombie_group), key=lambda z: z.x
    )
    patrol_bots_sorted: list[PatrolBot] = sorted(
        list(patrol_bot_group), key=lambda b: b.x
    )
    electrified_cells: set[tuple[int, int]] = set()
    if game_data.cell_size > 0:
        for bot in patrol_bots_sorted:
            if not bot.alive():
                continue
            cell = (
                int(bot.rect.centerx // game_data.cell_size),
                int(bot.rect.centery // game_data.cell_size),
            )
            if (
                0 <= cell[0] < game_data.layout.grid_cols
                and 0 <= cell[1] < game_data.layout.grid_rows
            ):
                electrified_cells.add(cell)
    game_data.state.electrified_cells = electrified_cells

    tracker_buckets: dict[tuple[int, int, int], list[Zombie]] = {}
    tracker_cell_size = ZOMBIE_TRACKER_CROWD_BAND_WIDTH
    angle_step = math.pi / 4.0
    zombie_kinds = (
        SpatialKind.ZOMBIE | SpatialKind.ZOMBIE_DOG | SpatialKind.TRAPPED_ZOMBIE
    )
    base_radius = ZOMBIE_SEPARATION_DISTANCE + PLAYER_SPEED
    for zombie in zombies_sorted:
        if (
            not zombie.alive()
            or not isinstance(zombie, Zombie)
            or zombie.kind != ZombieKind.TRACKER
        ):
            continue
        zombie.tracker_force_wander = False
        dx = zombie.last_move_dx
        dy = zombie.last_move_dy
        if abs(dx) <= 0.001 and abs(dy) <= 0.001:
            continue
        angle = math.atan2(dy, dx)
        angle_bin = int(round(angle / angle_step)) % 8
        cell_x = int(zombie.x // tracker_cell_size)
        cell_y = int(zombie.y // tracker_cell_size)
        tracker_buckets.setdefault((cell_x, cell_y, angle_bin), []).append(zombie)

    for bucket in tracker_buckets.values():
        if len(bucket) < ZOMBIE_TRACKER_GRID_CROWD_COUNT:
            continue
        RNG.choice(bucket).tracker_force_wander = True

    survivor_candidates = [
        survivor
        for survivor in survivor_group
        if survivor.alive() and not survivor.rescued
    ]
    dog_survivor_range_sq = (
        ZOMBIE_DOG_SURVIVOR_SIGHT_RANGE * ZOMBIE_DOG_SURVIVOR_SIGHT_RANGE
    )

    for zombie in zombies_sorted:
        target = target_center
        if getattr(zombie, "carbonized", False):
            zombie.on_moving_floor = False

        floor_dx, floor_dy = get_moving_floor_drift(
            zombie.rect,
            game_data.layout,
            cell_size=game_data.cell_size,
            speed=MOVING_FLOOR_SPEED,
        )
        zombie.on_moving_floor = abs(floor_dx) > 0.0 or abs(floor_dy) > 0.0

        if zombie.on_moving_floor and hasattr(zombie, "_apply_decay"):
            zombie._apply_decay()
            if not zombie.alive():
                continue
        if isinstance(zombie, ZombieDog):
            target = player.rect.center
            closest_survivor: Survivor | None = None
            closest_dist_sq = dog_survivor_range_sq
            for survivor in survivor_candidates:
                dx = survivor.rect.centerx - zombie.x
                dy = survivor.rect.centery - zombie.y
                dist_sq = dx * dx + dy * dy
                if dist_sq <= closest_dist_sq:
                    closest_survivor = survivor
                    closest_dist_sq = dist_sq
            if closest_survivor is not None:
                target = closest_survivor.rect.center
        if buddies_on_screen and not isinstance(zombie, ZombieDog):
            dist_to_target_sq = (target_center[0] - zombie.x) ** 2 + (
                target_center[1] - zombie.y
            ) ** 2
            nearest_buddy = min(
                buddies_on_screen,
                key=lambda buddy: (
                    (buddy.rect.centerx - zombie.x) ** 2
                    + (buddy.rect.centery - zombie.y) ** 2
                ),
            )
            dist_to_buddy_sq = (nearest_buddy.rect.centerx - zombie.x) ** 2 + (
                nearest_buddy.rect.centery - zombie.y
            ) ** 2
            if dist_to_buddy_sq < dist_to_target_sq:
                target = nearest_buddy.rect.center

        if stage.survivor_rescue_stage and not isinstance(zombie, ZombieDog):
            zombie_on_screen = rect_visible_on_screen(camera, zombie.rect)
            if zombie_on_screen:
                candidate_positions: list[tuple[int, int]] = []
                for survivor in survivors_on_screen:
                    candidate_positions.append(survivor.rect.center)
                for buddy in buddies_on_screen:
                    candidate_positions.append(buddy.rect.center)
                candidate_positions.append(player.rect.center)
                if candidate_positions:
                    target = min(
                        candidate_positions,
                        key=lambda pos: (
                            (pos[0] - zombie.x) ** 2 + (pos[1] - zombie.y) ** 2
                        ),
                    )
        if isinstance(zombie, ZombieDog):
            search_radius = max(ZOMBIE_DOG_PACK_CHASE_RANGE, base_radius)
        elif zombie.kind == ZombieKind.LINEFORMER:
            search_radius = max(ZOMBIE_LINEFORMER_JOIN_RADIUS, base_radius)
        else:
            search_radius = base_radius
        if zombie.kind == ZombieKind.SOLITARY and game_data.cell_size > 0:
            tile_x = int(zombie.x // game_data.cell_size)
            tile_y = int(zombie.y // game_data.cell_size)
            min_world_x = (tile_x - 1) * game_data.cell_size
            max_world_x = (tile_x + 2) * game_data.cell_size - 1
            min_world_y = (tile_y - 1) * game_data.cell_size
            max_world_y = (tile_y + 2) * game_data.cell_size - 1
            index_cell_size = max(1, int(spatial_index.cell_size))
            nearby_candidates = spatial_index.query_cells(
                min_cell_x=int(min_world_x // index_cell_size),
                max_cell_x=int(max_world_x // index_cell_size),
                min_cell_y=int(min_world_y // index_cell_size),
                max_cell_y=int(max_world_y // index_cell_size),
                kinds=zombie_kinds,
            )
        else:
            nearby_candidates = spatial_index.query_radius(
                (zombie.x, zombie.y),
                search_radius,
                kinds=zombie_kinds,
            )
        zombie_search_radius = (
            ZOMBIE_WALL_HUG_SENSOR_DISTANCE + zombie.collision_radius + 120
        )
        dog_candidates = nearby_candidates
        nearby_walls = _walls_near((zombie.x, zombie.y), zombie_search_radius)
        zombie.update(
            target,
            nearby_walls,
            dog_candidates,
            electrified_cells,
            footprints=game_data.state.footprints,
            cell_size=game_data.cell_size,
            layout=game_data.layout,
            now_ms=game_data.state.clock.elapsed_ms,
            drift=(floor_dx, floor_dy),
            spiky_plants=spiky_plants,
            trapped_spiky_plant_counts=trapped_spiky_plant_counts,
        )
        if not zombie.alive():
            last_damage_ms = getattr(zombie, "last_damage_ms", None)
            last_damage_source = getattr(zombie, "last_damage_source", None)
            died_from_damage = (
                last_damage_ms is not None and last_damage_ms == current_time
            )
            if died_from_damage and last_damage_source == "patrol_bot":
                died_from_damage = False
            if not died_from_damage:
                fov_target = active_car if player.in_car and active_car else player
                if rect_visible_on_screen(camera, zombie.rect) and is_entity_in_fov(
                    zombie.rect,
                    fov_target=fov_target,
                    flashlight_count=game_data.state.flashlight_count,
                ):
                    game_data.state.decay_effects.append(
                        DecayingEntityEffect(zombie.image, zombie.rect.center)
                    )
            continue

        if _is_on_fire_floor(
            zombie.x,
            zombie.y,
            cell_size=game_data.cell_size,
            fire_floor_cells=fire_floor_cells,
        ):
            zombie.kill()
            fov_target = active_car if player.in_car and active_car else player
            if rect_visible_on_screen(camera, zombie.rect) and is_entity_in_fov(
                zombie.rect,
                fov_target=fov_target,
                flashlight_count=game_data.state.flashlight_count,
            ):
                game_data.state.decay_effects.append(
                    DecayingEntityEffect(
                        zombie.image,
                        zombie.rect.center,
                        tone="burned",
                    )
                )
            continue

        # Check zombie pitfall
        pull_dist = zombie.collision_radius * 0.5
        pitfall_target_pos = pitfall_target(
            x=zombie.x,
            y=zombie.y,
            cell_size=game_data.cell_size,
            pitfall_cells=pitfall_cells,
            pull_distance=pull_dist,
        )
        if pitfall_target_pos is not None:
            zombie.kill()
            fall = FallingEntity(
                start_pos=(int(zombie.x), int(zombie.y)),
                target_pos=pitfall_target_pos,
                started_at_ms=game_data.state.clock.elapsed_ms,
                pre_fx_ms=0,
                fall_duration_ms=500,
                dust_duration_ms=0,
                kind=ZombieKind.DOG
                if isinstance(zombie, ZombieDog)
                else getattr(zombie, "kind", ZombieKind.NORMAL),
                mode="pitfall",
            )
            game_data.state.falling_zombies.append(fall)

    game_data.lineformer_trains.post_update(zombie_group)

    active_humans = [survivor for survivor in survivor_group if survivor.alive()]
    for bot in patrol_bots_sorted:
        if not bot.alive():
            continue
        floor_dx, floor_dy = get_moving_floor_drift(
            get_floor_overlap_rect(bot),
            game_data.layout,
            cell_size=game_data.cell_size,
            speed=MOVING_FLOOR_SPEED,
        )
        bot.on_moving_floor = abs(floor_dx) > 0.0 or abs(floor_dy) > 0.0
        bot_search_radius = bot.collision_radius + 120
        nearby_walls = _walls_near((bot.x, bot.y), bot_search_radius)
        bot.update(
            nearby_walls,
            patrol_bot_group=patrol_bot_group,
            human_group=active_humans,
            player=None if player.in_car else player,
            car=active_car,
            parked_cars=game_data.waiting_cars,
            cell_size=game_data.cell_size,
            pitfall_cells=pitfall_cells,
            fire_floor_cells=fire_floor_cells,
            layout=game_data.layout,
            drift=(floor_dx, floor_dy),
            now_ms=game_data.state.clock.elapsed_ms,
            spiky_plants=game_data.spiky_plants,
        )

    update_decay_effects(game_data.state.decay_effects, frames=1)



from __future__ import annotations

from typing import Any

import pygame
from pygame import surface

from ..colors import YELLOW, get_environment_palette
from ..models import FuelProgress, GameData
from ..render_assets import RenderAssets
from .entity_layer import _draw_entities, _draw_lineformer_train_markers
from .fog import _draw_fog_of_war, prewarm_fog_overlays
from .fx import _draw_decay_fx, _draw_fade_in_overlay, _draw_falling_fx
from .hud import (
    _build_objective_lines,
    _draw_endurance_timer,
    _draw_hint_indicator,
    _draw_inventory_icons,
    _draw_objective,
    _draw_status_bar,
    _draw_survivor_messages,
    _draw_timed_message,
    _draw_time_accel_indicator,
)
from .shadows import (
    _draw_wall_shadows,
    _get_shadow_layer,
    draw_entity_shadows_by_mode,
)
from .text_overlay import (
    blit_message,
    blit_message_wrapped,
    blit_text_wrapped,
    draw_pause_overlay,
    wrap_text,
)
from .world_tiles import _draw_footprints, _draw_play_area

__all__ = [
    "blit_message",
    "blit_message_wrapped",
    "blit_text_wrapped",
    "draw",
    "draw_pause_overlay",
    "prewarm_fog_overlays",
    "wrap_text",
]


def draw(
    assets: RenderAssets,
    screen: surface.Surface,
    game_data: GameData,
    *,
    config: dict[str, Any],
    hint_target: tuple[int, int] | None = None,
    contact_hint_targets: list[tuple[str, tuple[int, int]]] | None = None,
    hint_color: tuple[int, int, int] | None = None,
    fps: float | None = None,
) -> None:
    hint_color = hint_color or YELLOW
    state = game_data.state
    player = game_data.player
    if player is None:
        raise ValueError("draw requires an active player on game_data")

    camera = game_data.camera
    stage = game_data.stage
    outside_cells = game_data.layout.outside_cells
    all_sprites = game_data.groups.all_sprites
    has_fuel = state.fuel_progress == FuelProgress.FULL_CAN
    has_empty_fuel_can = state.fuel_progress == FuelProgress.EMPTY_CAN
    flashlight_count = state.flashlight_count
    active_car = game_data.car if game_data.car and game_data.car.alive() else None
    if player.in_car and game_data.car and game_data.car.alive():
        fov_target = game_data.car
    else:
        fov_target = player

    palette = get_environment_palette(state.ambient_palette_key)
    screen.fill(palette.outside)

    _draw_play_area(
        screen,
        camera.apply_rect,
        pygame.Rect(
            -camera.camera.x,
            -camera.camera.y,
            assets.screen_width,
            assets.screen_height,
        ),
        assets,
        palette,
        game_data.layout.field_rect,
        outside_cells,
        game_data.layout.fall_spawn_cells,
        game_data.layout.pitfall_cells,
        game_data.layout.fire_floor_cells,
        game_data.layout.metal_floor_cells,
        game_data.layout.puddle_cells,
        game_data.layout.moving_floor_cells,
        state.electrified_cells,
        game_data.cell_size,
        elapsed_ms=int(state.clock.elapsed_ms),
    )
    shadows_enabled = config.get("visual", {}).get("shadows", {}).get("enabled", True)
    if shadows_enabled:
        dawn_shadow_mode = bool(stage and stage.endurance_stage and state.dawn_ready)
        lsp = (
            None
            if dawn_shadow_mode
            else (fov_target.rect.center if fov_target else None)
        )
        light_source_pos: tuple[float, float] | None = (
            (float(lsp[0]), float(lsp[1])) if lsp is not None else None
        )
        shadow_layer = _get_shadow_layer(screen.get_size())
        shadow_layer.fill((0, 0, 0, 0))
        drew_shadow = _draw_wall_shadows(
            shadow_layer,
            camera.apply_rect,
            wall_cells=game_data.layout.wall_cells,
            steel_beam_cells=game_data.layout.steel_beam_cells,
            outer_wall_cells=game_data.layout.outer_wall_cells,
            cell_size=game_data.cell_size,
            light_source_pos=light_source_pos,
        )
        drew_shadow |= draw_entity_shadows_by_mode(
            shadow_layer,
            camera.apply_rect,
            all_sprites,
            dawn_shadow_mode=dawn_shadow_mode,
            light_source_pos=light_source_pos,
            outside_cells=outside_cells,
            cell_size=game_data.cell_size,
            flashlight_count=flashlight_count,
        )
        if drew_shadow:
            screen.blit(shadow_layer, (0, 0))
    _draw_footprints(
        screen,
        camera.apply_rect,
        assets,
        state.footprints,
        config=config,
        now_ms=state.clock.elapsed_ms,
    )
    _draw_entities(
        screen,
        [(entity, camera.apply_rect(entity.rect)) for entity in all_sprites],
        player,
        has_fuel=has_fuel,
        has_empty_fuel_can=has_empty_fuel_can,
        show_fuel_indicator=not (stage and stage.endurance_stage),
    )
    marker_draw_data_screen = []
    for (
        world_x,
        world_y,
        angle_rad,
    ) in game_data.lineformer_trains.iter_marker_draw_data(
        game_data.groups.zombie_group
    ):
        world_center = pygame.Rect(
            int(round(world_x)),
            int(round(world_y)),
            0,
            0,
        )
        center_x, center_y = camera.apply_rect(world_center).topleft
        marker_draw_data_screen.append((center_x, center_y, angle_rad))
    _draw_lineformer_train_markers(
        screen,
        marker_draw_data_screen,
    )

    _draw_decay_fx(
        screen,
        camera.apply_rect,
        state.decay_effects,
    )

    _draw_falling_fx(
        screen,
        camera.apply_rect,
        state.falling_zombies,
        state.flashlight_count,
        state.dust_rings,
        state.clock.elapsed_ms,
    )

    _draw_hint_indicator(
        screen,
        camera,
        assets,
        player,
        hint_target,
        contact_hint_targets=contact_hint_targets,
        hint_color=hint_color,
        flashlight_count=flashlight_count,
    )
    fov_center_screen: tuple[int, int] | None = None
    if fov_target is not None:
        fov_center_screen = tuple(map(int, camera.apply(fov_target).center))
    _draw_fog_of_war(
        screen,
        assets,
        game_data.fog,
        fov_center_screen,
        stage=stage,
        flashlight_count=flashlight_count,
        dawn_ready=state.dawn_ready,
    )

    objective_lines = _build_objective_lines(
        stage=stage,
        state=state,
        player=player,
        active_car=active_car,
        fuel_progress=state.fuel_progress,
        buddy_merged_count=state.buddy_merged_count,
        buddy_required=stage.buddy_required_count if stage else 0,
    )
    if objective_lines:
        _draw_objective(objective_lines, screen=screen)
    _draw_inventory_icons(
        screen,
        assets,
        has_fuel=has_fuel,
        has_empty_fuel_can=has_empty_fuel_can,
        flashlight_count=flashlight_count,
        shoes_count=state.shoes_count,
        player_in_car=player.in_car,
        buddy_onboard=state.buddy_onboard,
        survivors_onboard=state.survivors_onboard,
        passenger_capacity=state.survivor_capacity,
    )
    _draw_survivor_messages(screen, assets, list(state.survivor_messages))
    _draw_endurance_timer(screen, assets, stage=stage, state=state)
    _draw_time_accel_indicator(screen, assets, stage=stage)
    _draw_status_bar(
        screen,
        assets,
        config,
        stage=stage,
        seed=state.seed,
        debug_mode=state.debug_mode,
        zombie_group=game_data.groups.zombie_group,
        lineformer_marker_count=game_data.lineformer_trains.total_marker_count(),
        falling_spawn_carry=state.falling_spawn_carry,
        show_fps=state.show_fps,
        fps=fps,
    )

    _draw_fade_in_overlay(screen, state)
    _draw_timed_message(
        screen,
        assets,
        message=state.timed_message,
        elapsed_play_ms=state.clock.elapsed_ms,
    )



"""Level layout constants."""

from __future__ import annotations

DEFAULT_GRID_COLS = 48
DEFAULT_GRID_ROWS = 30
DEFAULT_CELL_SIZE = 50  # world units per cell; adjust to scale the whole map
DEFAULT_WALL_LINES = 80  # reduced density (roughly 1/5 of previous 450)
DEFAULT_GRID_WIRE_WALL_LINES = int(DEFAULT_WALL_LINES * 0.7)
DEFAULT_SPARSE_WALL_DENSITY = 0.10
DEFAULT_CORRIDOR_BREAK_RATIO = 0.03
DEFAULT_STEEL_BEAM_CHANCE = 0.02

__all__ = [
    "DEFAULT_GRID_COLS",
    "DEFAULT_GRID_ROWS",
    "DEFAULT_CELL_SIZE",
    "DEFAULT_WALL_LINES",
    "DEFAULT_GRID_WIRE_WALL_LINES",
    "DEFAULT_SPARSE_WALL_DENSITY",
    "DEFAULT_CORRIDOR_BREAK_RATIO",
    "DEFAULT_STEEL_BEAM_CHANCE",
]



from __future__ import annotations
from enum import Enum
from typing import Any, Callable

import numpy as np  # type: ignore
import pygame
import pygame.surfarray as pg_surfarray  # type: ignore
from pygame import surface

from ..gameplay_constants import DEFAULT_FLASHLIGHT_SPAWN_COUNT
from ..models import Stage
from ..render_assets import RenderAssets
from .hud import _get_fog_scale


def _build_bayer_matrix(size: int) -> np.ndarray:
    """Build an NxN Bayer threshold matrix where N is a power of two."""
    if size < 2 or (size & (size - 1)) != 0:
        raise ValueError("Bayer matrix size must be a power of two >= 2")
    matrix = np.array([[0, 2], [3, 1]], dtype=np.int32)
    while matrix.shape[0] < size:
        matrix = np.block(
            [
                [4 * matrix + 0, 4 * matrix + 2],
                [4 * matrix + 3, 4 * matrix + 1],
            ]
        )
    return matrix


_BAYER_MATRIX_16 = _build_bayer_matrix(16)


def _max_flashlight_pickups() -> int:
    """Return the maximum flashlight pickups available per stage."""
    return max(1, DEFAULT_FLASHLIGHT_SPAWN_COUNT)


class _FogProfile(Enum):
    DARK0 = (0, (0, 0, 0, 255))
    DARK1 = (1, (0, 0, 0, 255))
    DARK2 = (2, (0, 0, 0, 255))

    def __init__(self, flashlight_count: int, color: tuple[int, int, int, int]) -> None:
        self.flashlight_count = flashlight_count
        self.color = color

    def _scale(self, assets: RenderAssets, _stage: Stage | None) -> float:
        count = max(0, min(self.flashlight_count, _max_flashlight_pickups()))
        return _get_fog_scale(assets, count)

    @staticmethod
    def _from_flashlight_count(count: int) -> "_FogProfile":
        safe_count = max(0, count)
        if safe_count >= 2:
            return _FogProfile.DARK2
        if safe_count == 1:
            return _FogProfile.DARK1
        return _FogProfile.DARK0


def prewarm_fog_overlays(
    fog_data: dict[str, Any],
    assets: RenderAssets,
    *,
    stage: Stage | None = None,
    progress_callback: Callable[[int, int], None] | None = None,
) -> None:
    """Populate fog overlay cache for each reachable flashlight count."""

    profiles = list(_FogProfile)
    total = len(profiles)
    for index, profile in enumerate(profiles, start=1):
        _get_fog_overlay_surfaces(
            fog_data,
            assets,
            profile,
            stage=stage,
        )
        if progress_callback is not None:
            progress_callback(index, total)


def _soften_surface(
    source: surface.Surface,
    scale: float,
) -> surface.Surface:
    """Return a softened copy of a surface using smoothscale down/up."""
    safe_scale = max(0.5, min(1.0, scale))
    if safe_scale >= 0.999:
        return source
    width, height = source.get_size()
    softened_size = (max(1, int(width * safe_scale)), max(1, int(height * safe_scale)))
    softened = pygame.transform.smoothscale(source, softened_size)
    return pygame.transform.smoothscale(softened, (width, height))


def _build_continuous_hatch_surface(
    size: tuple[int, int],
    center: tuple[int, int],
    base_color: tuple[int, int, int, int],
    max_radius: float,
    density_ramps: list[tuple[float, float]],
    spacing: int = 4,
) -> surface.Surface:
    width, height = size
    hatch = pygame.Surface((width, height), pygame.SRCALPHA)
    hatch.fill((base_color[0], base_color[1], base_color[2], 0))

    bayer = _BAYER_MATRIX_16
    ramps: list[tuple[float, float, float]] = []
    for ramp_start_ratio, ramp_max_density in density_ramps:
        safe_start = max(0.0, min(1.0, ramp_start_ratio))
        safe_max_density = max(0.0, min(1.0, ramp_max_density))
        start_radius = max(0.0, max_radius * safe_start)
        fade_range = max(1.0, max_radius - start_radius)
        ramps.append((safe_max_density, start_radius, fade_range))
    base_alpha = base_color[3]
    cell_center = (spacing - 1) / 2.0

    cx, cy = center
    alpha_view = pg_surfarray.pixels_alpha(hatch)
    yy, xx = np.indices((height, width))
    dx = xx - cx
    dy = yy - cy
    radius = np.hypot(dx, dy)
    density = np.zeros_like(radius)
    for ramp_max_density, ramp_start, ramp_range in ramps:
        progress = (radius - ramp_start) / ramp_range
        progress = np.clip(progress, 0.0, 1.0)
        density += ramp_max_density * progress
    density = np.clip(density, 0.0, 1.0)
    bayer_size = bayer.shape[0]
    threshold = (density * (bayer_size * bayer_size)).astype(np.int32)
    grid_x = (xx // spacing) % bayer_size
    grid_y = (yy // spacing) % bayer_size
    bayer_vals = bayer[grid_y, grid_x]
    mask = bayer_vals < threshold

    dot_radius = np.maximum(1.0, density * spacing)
    local_x = (xx % spacing) - cell_center
    local_y = (yy % spacing) - cell_center
    mask &= (local_x * local_x + local_y * local_y) <= (dot_radius * dot_radius)

    alpha_view[:, :] = (mask.T * base_alpha).astype(np.uint8)
    del alpha_view

    return hatch


def _build_edge_feather_surface(
    size: tuple[int, int],
    center: tuple[int, int],
    max_radius: int,
    softness_px: int,
) -> surface.Surface:
    """Return a radial alpha feather around the FOV edge."""
    width, height = size
    feather = pygame.Surface((width, height), pygame.SRCALPHA)
    feather.fill((0, 0, 0, 0))
    if softness_px <= 0:
        return feather

    inner_radius = max(0.0, float(max_radius - softness_px))
    outer_radius = float(max_radius + softness_px)
    transition = max(1.0, outer_radius - inner_radius)

    alpha_view = pg_surfarray.pixels_alpha(feather)
    yy, xx = np.indices((height, width))
    dx = xx - center[0]
    dy = yy - center[1]
    dist = np.hypot(dx, dy)
    progress = np.clip((dist - inner_radius) / transition, 0.0, 1.0)
    alpha_view[:, :] = (progress.T * 255).astype(np.uint8)
    del alpha_view
    return feather


def _get_fog_overlay_surfaces(
    fog_data: dict[str, Any],
    assets: RenderAssets,
    profile: _FogProfile,
    *,
    stage: Stage | None = None,
) -> dict[str, Any]:
    overlays = fog_data.setdefault("overlays", {})
    key = profile
    if key in overlays:
        return overlays[key]

    scale = profile._scale(assets, stage)
    max_radius = int(assets.fov_radius * scale)
    padding = 32
    coverage_width = max(assets.screen_width * 2, max_radius * 2)
    coverage_height = max(assets.screen_height * 2, max_radius * 2)
    width = coverage_width + padding * 2
    height = coverage_height + padding * 2

    aa_scale = max(1, int(assets.fog_layer_aa_scale))
    render_width = max(1, width * aa_scale)
    render_height = max(1, height * aa_scale)
    render_center = (render_width // 2, render_height // 2)
    render_radius = max(1, int(max_radius * aa_scale))

    hard_surface_render = pygame.Surface((render_width, render_height), pygame.SRCALPHA)
    base_color = profile.color
    hard_surface_render.fill(base_color)
    pygame.draw.circle(hard_surface_render, (0, 0, 0, 0), render_center, render_radius)

    ring_surfaces: list[surface.Surface] = []

    combined_surface_render = hard_surface_render.copy()
    edge_feather = _build_edge_feather_surface(
        (render_width, render_height),
        render_center,
        render_radius,
        max(0, int(assets.fog_edge_softness_px * aa_scale)),
    )
    combined_surface_render.blit(edge_feather, (0, 0))
    hatch_surface = _build_continuous_hatch_surface(
        (render_width, render_height),
        render_center,
        base_color,
        render_radius,
        assets.fog_hatch_density_ramps,
        spacing=max(1, 4 * aa_scale),
    )
    if assets.fog_hatch_soften_scale < 1.0:
        hatch_surface = _soften_surface(
            hatch_surface,
            assets.fog_hatch_soften_scale,
        )
    combined_surface_render.blit(hatch_surface, (0, 0))

    visible_fade_surface = _build_flashlight_fade_surface(
        (render_width, render_height),
        render_center,
        render_radius,
        outer_extension=max(1, 30 * aa_scale),
    )
    combined_surface_render.blit(visible_fade_surface, (0, 0))
    if aa_scale > 1:
        hard_surface = pygame.transform.smoothscale(hard_surface_render, (width, height))
        combined_surface = pygame.transform.smoothscale(
            combined_surface_render, (width, height)
        )
    else:
        hard_surface = hard_surface_render
        combined_surface = combined_surface_render

    overlay_entry = {
        "hard": hard_surface,
        "rings": ring_surfaces,
        "combined": combined_surface,
    }
    overlays[key] = overlay_entry
    return overlay_entry


def _build_flashlight_fade_surface(
    size: tuple[int, int],
    center: tuple[int, int],
    max_radius: int,
    *,
    start_ratio: float = 0.2,
    max_alpha: int = 220,
    outer_extension: int = 30,
) -> surface.Surface:
    """Return a radial gradient so flashlight edges softly darken again."""

    width, height = size
    fade_surface = pygame.Surface((width, height), pygame.SRCALPHA)
    fade_surface.fill((0, 0, 0, 0))

    start_radius = max(0.0, min(max_radius, max_radius * start_ratio))
    end_radius = max(start_radius + 1, max_radius + outer_extension)
    fade_range = max(1.0, end_radius - start_radius)

    cx, cy = center
    yy, xx = np.indices((height, width))
    dx = xx - cx
    dy = yy - cy
    dist = np.hypot(dx, dy)
    dist = np.minimum(dist, end_radius)
    progress = np.clip((dist - start_radius) / fade_range, 0.0, 1.0)
    alpha = (progress * max_alpha).astype(np.uint8)
    alpha_view = pg_surfarray.pixels_alpha(fade_surface)
    alpha_view[:, :] = alpha.T
    del alpha_view

    return fade_surface


def _draw_fog_of_war(
    screen: surface.Surface,
    assets: RenderAssets,
    fog_surfaces: dict[str, Any],
    fov_center_screen: tuple[int, int] | None,
    *,
    stage: Stage | None,
    flashlight_count: int,
    dawn_ready: bool,
) -> None:
    if fov_center_screen is None:
        return
    if stage and stage.endurance_stage and dawn_ready:
        return
    profile = _FogProfile._from_flashlight_count(flashlight_count)
    overlay = _get_fog_overlay_surfaces(
        fog_surfaces,
        assets,
        profile,
        stage=stage,
    )
    combined_surface: surface.Surface = overlay["combined"]
    screen.blit(
        combined_surface,
        combined_surface.get_rect(center=fov_center_screen),
    )



from __future__ import annotations

from typing import cast

import pygame

from ..world_grid import WallIndex, walls_for_radius
from .movement import _circle_wall_collision
from .walls import Wall


def _sprite_center_and_radius(
    sprite: pygame.sprite.Sprite,
) -> tuple[tuple[int, int], float]:
    center = sprite.rect.center
    if hasattr(sprite, "radius"):
        radius = float(getattr(sprite, "collision_radius", sprite.radius))
    else:
        radius = float(max(sprite.rect.width, sprite.rect.height) / 2)
    return center, radius


def _sprite_collision_circle(
    sprite: pygame.sprite.Sprite,
) -> tuple[tuple[int, int], float]:
    if hasattr(sprite, "get_collision_circle"):
        center, radius = sprite.get_collision_circle()
        return (int(center[0]), int(center[1])), float(radius)
    return _sprite_center_and_radius(sprite)


def collide_circle_custom(
    sprite_a: pygame.sprite.Sprite, sprite_b: pygame.sprite.Sprite
) -> bool:
    center_a, radius_a = _sprite_collision_circle(sprite_a)
    center_b, radius_b = _sprite_collision_circle(sprite_b)
    dx = center_a[0] - center_b[0]
    dy = center_a[1] - center_b[1]
    radius = radius_a + radius_b
    return dx * dx + dy * dy <= radius * radius


def _walls_for_sprite(
    sprite: pygame.sprite.Sprite,
    wall_index: WallIndex,
    *,
    cell_size: int,
    grid_cols: int,
    grid_rows: int,
) -> list[Wall]:
    center, radius = _sprite_center_and_radius(sprite)
    return walls_for_radius(
        wall_index,
        center,
        radius,
        cell_size=cell_size,
        grid_cols=grid_cols,
        grid_rows=grid_rows,
    )


def _collide_sprite_wall(
    sprite: pygame.sprite.Sprite, wall: pygame.sprite.Sprite
) -> bool:
    if hasattr(sprite, "radius"):
        center = sprite.rect.center
        radius = float(getattr(sprite, "collision_radius", sprite.radius))
        return _circle_wall_collision(center, radius, wall)
    if hasattr(wall, "collides_rect"):
        return wall.collides_rect(sprite.rect)
    if hasattr(sprite, "collides_rect"):
        return sprite.collides_rect(wall.rect)
    return sprite.rect.colliderect(wall.rect)


def spritecollideany_walls(
    sprite: pygame.sprite.Sprite,
    walls: pygame.sprite.Group,
    *,
    wall_index: WallIndex | None = None,
    cell_size: int | None = None,
    grid_cols: int | None = None,
    grid_rows: int | None = None,
) -> Wall | None:
    if wall_index is None:
        return cast(
            Wall | None,
            pygame.sprite.spritecollideany(
                sprite, walls, collided=_collide_sprite_wall
            ),
        )
    if cell_size is None or grid_cols is None or grid_rows is None:
        raise ValueError(
            "cell_size/grid_cols/grid_rows are required when using wall_index"
        )
    for wall in _walls_for_sprite(
        sprite,
        wall_index,
        cell_size=cell_size,
        grid_cols=grid_cols,
        grid_rows=grid_rows,
    ):
        if _collide_sprite_wall(sprite, wall):
            return wall
    return None



from __future__ import annotations

from contextlib import contextmanager
from importlib import resources
from pathlib import Path
from typing import Iterator

import pygame

_FONT_CACHE: dict[tuple[str | None, int], pygame.font.Font] = {}
_FONT_META: dict[int, tuple[str | None, int]] = {}


@contextmanager
def _resource_path(resource: str | None) -> Iterator[Path | None]:
    if not resource:
        yield None
        return

    try:
        font = resources.files("zombie_escape")
        for part in Path(resource).parts:
            font = font.joinpath(part)
        with resources.as_file(font) as path:
            yield path
    except (FileNotFoundError, ModuleNotFoundError):
        yield None


def load_font(resource: str | None, size: int) -> pygame.font.Font:
    """Load and cache a pygame font for the given resource and size."""
    normalized_size = max(1, int(size))
    cache_key = (resource, normalized_size)
    cached = _FONT_CACHE.get(cache_key)
    if cached is not None:
        return cached

    with _resource_path(resource) as path:
        font = pygame.font.Font(str(path) if path else None, normalized_size)

    _FONT_CACHE[cache_key] = font
    _FONT_META[id(font)] = (resource, normalized_size)
    return font


def render_text_surface(
    font: pygame.font.Font,
    text: str,
    color: tuple[int, int, int],
    *,
    line_height_scale: float = 1.0,
) -> pygame.Surface:
    surface = font.render(text, False, color)
    line_height = int(round(font.get_linesize() * max(0.0, line_height_scale)))
    if line_height <= 0 or line_height <= surface.get_height():
        return surface
    padded = pygame.Surface((surface.get_width(), line_height), pygame.SRCALPHA)
    top_pad = max(0, (line_height - surface.get_height()) // 2)
    padded.blit(surface, (0, top_pad))
    return padded


def clear_font_cache() -> None:
    _FONT_CACHE.clear()
    _FONT_META.clear()



"""Screen and window related constants."""

from __future__ import annotations

SCREEN_WIDTH = 400  # Base logical render width
SCREEN_HEIGHT = 300  # Base logical render height
DEFAULT_WINDOW_SCALE = 2.0  # Keep ~800x600 OS window while rendering at 400x300
WINDOW_SCALE_MIN = 1.0
WINDOW_SCALE_MAX = DEFAULT_WINDOW_SCALE * 2  # Allow up to 1600x1200 windows
FPS = 60
STATUS_BAR_HEIGHT = 18

__all__ = [
    "SCREEN_WIDTH",
    "SCREEN_HEIGHT",
    "DEFAULT_WINDOW_SCALE",
    "WINDOW_SCALE_MIN",
    "WINDOW_SCALE_MAX",
    "FPS",
    "STATUS_BAR_HEIGHT",
]
